# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import copy
from functools import partial

import pytest

from awscli.autocomplete import parser
from awscli.testutils import unittest
from tests.unit.autocomplete import InMemoryIndex

# This models an 'aws ec2 stop-instances' command
# along with the 'region', 'endpoint-url', and 'debug' global params.
SAMPLE_MODEL = InMemoryIndex(
    # This format is intended to match the structure you'd get from
    # querying the index db to minimize any parity issues between
    # the tests and the real indexer.
    index={
        'arg_names': {
            '': {
                'aws': ['region', 'endpoint-url', 'debug'],
            },
            'aws.ec2': {
                'stop-instances': ['instance-ids', 'foo-arg'],
            },
            'aws.s3': {
                'cp': ['bucket', 'key'],
            },
            'aws.s3api': {
                'get-object': ['outfile', 'bucket', 'key'],
            },
            'aws.logs': {'tail': ['group_name', 'filter-pattern']},
        },
        'command_names': {
            '': [('aws', None)],
            'aws': [
                ('ec2', None),
                ('logs', None),
                ('s3api', None),
                ('s3', None),
            ],
            'aws.ec2': [('stop-instances', None)],
            'aws.logs': [('tail', None)],
            'aws.s3': [('cp', None)],
            'aws.s3api': [('get-object', None)],
        },
        'arg_data': {
            '': {
                'aws': {
                    'debug': (
                        'debug',
                        'boolean',
                        'aws',
                        '',
                        None,
                        False,
                        False,
                    ),
                    'endpoint-url': (
                        'endpoint-url',
                        'string',
                        'aws',
                        '',
                        None,
                        False,
                        False,
                    ),
                    'region': (
                        'region',
                        'string',
                        'aws',
                        '',
                        None,
                        False,
                        False,
                    ),
                }
            },
            'aws.ec2': {
                'stop-instances': {
                    'instance-ids': (
                        'instance-ids',
                        'string',
                        'stop-instances',
                        'aws.ec2.',
                        '*',
                        False,
                        False,
                    ),
                    'foo-arg': (
                        'foo-arg',
                        'string',
                        'stop-instances',
                        'aws.ec2',
                        None,
                        False,
                        False,
                    ),
                    'positional': (
                        'positional',
                        'string',
                        'stop-instances',
                        'aws.ec2',
                        None,
                        True,
                        False,
                    ),
                }
            },
            'aws.logs': {
                'tail': {
                    'group_name': (
                        'group_name',
                        'string',
                        'tail',
                        'aws.logs.',
                        None,
                        True,
                        False,
                    ),
                    'filter-pattern': (
                        'filter-pattern',
                        'string',
                        'tail',
                        'aws.logs',
                        None,
                        False,
                        False,
                    ),
                }
            },
            'aws.s3': {
                'cp': {
                    'bucket': (
                        'bucket',
                        'string',
                        'cp',
                        'aws.s3.',
                        None,
                        False,
                        False,
                    ),
                    'key': (
                        'key',
                        'string',
                        'cp',
                        'aws.s3.',
                        None,
                        False,
                        False,
                    ),
                }
            },
            'aws.s3api': {
                'get-object': {
                    'outfile': (
                        'outfile',
                        'string',
                        'get-object',
                        'aws.s3api.',
                        None,
                        False,
                        False,
                    ),
                    'bucket': (
                        'bucket',
                        'string',
                        'get-object',
                        'aws.s3api.',
                        None,
                        False,
                        False,
                    ),
                    'key': (
                        'key',
                        'string',
                        'get-object',
                        'aws.s3api.',
                        None,
                        False,
                        False,
                    ),
                }
            },
        },
    },
)


@pytest.mark.parametrize(
    'command_line',
    [
        (
            'aws ec2 stop-instances '
            '--instance-ids i-123 i-124 --foo-arg value --debug '
            '--endpoint-url https://foo '
        ),
        (
            'aws --debug ec2 stop-instances '
            '--instance-ids i-123 i-124 --foo-arg value '
            '--endpoint-url https://foo '
        ),
        (
            'aws --endpoint-url https://foo --debug ec2 stop-instances '
            '--instance-ids i-123 i-124 --foo-arg value '
        ),
        (
            'aws ec2 --debug --endpoint-url https://foo stop-instances '
            '--instance-ids i-123 i-124 --foo-arg value '
        ),
        (
            'aws ec2 stop-instances --debug --endpoint-url https://foo '
            '--instance-ids i-123 i-124 --foo-arg value '
        ),
        (
            'aws ec2 --endpoint-url https://foo stop-instances --debug '
            '--instance-ids i-123 i-124 --foo-arg value '
        ),
    ],
)
def test_can_handle_arbitrary_ordering(command_line):
    # This test verifies that parse to the same result given a command
    # with its arguments in varying orders.  For more detailed testing,
    # see TestCanParseCLICommand below.
    expected = parser.ParsedResult(
        current_command='stop-instances',
        global_params={'debug': True, 'endpoint-url': 'https://foo'},
        parsed_params={'instance-ids': ['i-123', 'i-124'], 'foo-arg': 'value'},
        lineage=['aws', 'ec2'],
        current_fragment='',
    )
    _assert_parses_to(command_line, expected)


def _assert_parses_to(command_line, expected):
    p = parser.CLIParser(SAMPLE_MODEL)
    result = p.parse(command_line)
    for key, value in vars(expected).items():
        actual = getattr(result, key)
        assert getattr(result, key) == value, '%r != %r for attribute: %r' % (
            actual,
            value,
            key,
        )


def _generate_command_chunks():
    command_line = (
        'aws ec2 stop-instances --instance-ids i-123 i-124 '
        '--foo-arg value --debug --endpoint-url https://foo'
    )
    return [command_line[:i] for i in range(1, len(command_line))]


@pytest.mark.parametrize("chunk", _generate_command_chunks())
def test_properties_of_unparsed_results(chunk):
    # The parser should never raise an exception.  If it can't
    # understand something it should still return a ParsedResult
    # with the parts it doesn't understand addded to the unparsed_items
    # attribute.  The ParsedResult should always have some basic invariants
    # we can verify which are called out in the tests below.
    # This test ensures that at every single slice of the full command_line
    # we always produce a sensical ParsedResult.
    cli_parser = parser.CLIParser(SAMPLE_MODEL)
    _assert_parsed_properties(chunk, cli_parser)


def _assert_parsed_properties(chunk, cli_parser):
    result = cli_parser.parse(chunk)
    assert isinstance(result, parser.ParsedResult)
    if chunk[-1].isspace():
        # If there's a space as the last char, then we should have
        # a current_fragment of an empty string.  This results in
        # all results being returned from the prefix match in the
        # auto-completer.
        assert result.current_fragment == ''
    elif result.current_fragment is not None:
        # The current_fragment, if not None is always the last part
        # of the command line.
        assert chunk.endswith(result.current_fragment)


class TestCanParseCLICommand(unittest.TestCase):
    def setUp(self):
        self.cli_parser = self.create_parser()

    def create_parser(self, return_first_command_match=False):
        return parser.CLIParser(
            SAMPLE_MODEL, return_first_command_match=return_first_command_match
        )

    def assert_parsed_results_equal(self, actual, **expected):
        # Asserts that every kwargs in expected matches what was actually
        # parsed.
        for key, value in expected.items():
            actual_value = getattr(actual, key)
            self.assertEqual(
                actual_value,
                value,
                '%r != %r (attr: %r)' % (actual_value, value, key),
            )

    def assert_parsed_s3api_result_correct(self, result):
        self.assert_parsed_results_equal(
            result,
            current_command='get-object',
            parsed_params={
                'outfile': 'filename',
                'bucket': 'myBucket',
                'key': 'foo',
            },
            lineage=['aws', 's3api'],
            unparsed_items=[],
        )

    def test_parsed_result_not_equal(self):
        self.assertFalse(parser.ParsedResult(current_command='ec2') == 'ec2')

    def test_can_handle_empty_string_for_command(self):
        result = self.cli_parser.parse('')
        self.assert_parsed_results_equal(
            result,
            current_command=None,
            global_params={},
            parsed_params={},
            lineage=[],
            unparsed_items=[],
        )

    def test_result_if_aws_exe_alias(self):
        result = self.cli_parser.parse('aws2 ec2')
        self.assert_parsed_results_equal(
            result,
            current_command='ec2',
            global_params={},
            parsed_params={},
            lineage=['aws'],
            unparsed_items=[],
        )

    def test_can_parse_operation_command_accepts_single_value_arg(self):
        result = self.cli_parser.parse('aws ec2 stop-instances --foo-arg bar ')
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            global_params={},
            parsed_params={'foo-arg': 'bar'},
            lineage=['aws', 'ec2'],
        )

    def test_can_parse_operation_command_with_param(self):
        result = self.cli_parser.parse(
            'aws ec2 stop-instances --instance-ids i-1 '
        )
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            global_params={},
            parsed_params={'instance-ids': ['i-1']},
            lineage=['aws', 'ec2'],
        )

    def test_can_parse_bool_param(self):
        result = self.cli_parser.parse(
            'aws --debug ec2 stop-instances --instance-ids i-1 '
        )
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            global_params={'debug': True},
            parsed_params={'instance-ids': ['i-1']},
            lineage=['aws', 'ec2'],
        )

    def test_can_parse_bool_param_in_any_location(self):
        result = self.cli_parser.parse(
            'aws ec2 stop-instances --instance-ids i-1 --debug'
        )
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            global_params={'debug': True},
            parsed_params={'instance-ids': ['i-1']},
            lineage=['aws', 'ec2'],
        )

    def test_can_parse_operation_command(self):
        result = self.cli_parser.parse('aws ec2 stop-instances')
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            global_params={},
            parsed_params={},
            lineage=['aws', 'ec2'],
        )

    def test_can_parse_service_command(self):
        result = self.cli_parser.parse('aws ec2')
        self.assert_parsed_results_equal(
            result,
            current_command='ec2',
            parsed_params={},
            lineage=['aws'],
        )

    def test_can_parse_aws_command(self):
        result = self.cli_parser.parse('aws')
        self.assert_parsed_results_equal(
            result,
            current_command='aws',
            parsed_params={},
            global_params={},
            lineage=[],
        )

    def test_ignores_unknown_args(self):
        result = self.cli_parser.parse(
            'aws ec2 stop-instances --unknown-arg bar'
        )
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            parsed_params={},
            lineage=['aws', 'ec2'],
        )

    def test_can_parse_if_there_longer_command(self):
        result = self.cli_parser.parse('aws s3')
        self.assert_parsed_results_equal(
            result,
            current_command='aws',
            parsed_params={},
            global_params={},
            current_fragment='s3',
            lineage=[],
        )

    def test_parse_shortest_command_if_first_wins(self):
        parser = self.create_parser(return_first_command_match=True)
        result = parser.parse('aws s3')
        self.assert_parsed_results_equal(
            result,
            current_command='s3',
            parsed_params={},
            global_params={},
            lineage=['aws'],
        )

    def test_can_proceed_if_there_longer_command(self):
        result = self.cli_parser.parse('aws s3 ')
        self.assert_parsed_results_equal(
            result,
            current_command='s3',
            global_params={},
            current_fragment='',
            lineage=['aws'],
        )

    def test_can_consume_one_or_more_nargs(self):
        result = self.cli_parser.parse(
            'aws ec2 stop-instances --instance-ids i-1 i-2 i-3 '
        )
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            parsed_params={'instance-ids': ['i-1', 'i-2', 'i-3']},
            lineage=['aws', 'ec2'],
        )

    def test_can_consume_zero_or_one_nargs(self):
        model = copy.deepcopy(SAMPLE_MODEL)
        nargs_one_or_more = '?'
        model.index['arg_data']['aws.ec2']['stop-instances']['foo-arg'] = (
            'foo-arg',
            'string',
            'stop-instances',
            'aws.ec2',
            nargs_one_or_more,
            False,
            False,
        )
        p = parser.CLIParser(model)
        self.assertEqual(
            p.parse('aws ec2 stop-instances --foo-arg --debug').parsed_params[
                'foo-arg'
            ],
            None,
        )
        self.assertEqual(
            p.parse(
                'aws ec2 stop-instances --foo-arg bar --debug'
            ).parsed_params['foo-arg'],
            'bar',
        )

    def test_truncates_line_based_on_location(self):
        # The 22nd index cuts off right after `stop-instances`.
        result = self.cli_parser.parse(
            'aws ec2 stop-instances --instance-ids i-1 i-2 i-3', 22
        )
        # We should not have parsed the 'instance-ids'.
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            parsed_params={},
            lineage=['aws', 'ec2'],
        )

    def test_preserves_current_word(self):
        result = self.cli_parser.parse('aws ec2 stop-')
        self.assert_parsed_results_equal(
            result,
            current_command='ec2',
            parsed_params={},
            global_params={},
            lineage=['aws'],
            current_fragment='stop-',
        )

    def test_word_not_preserved_if_not_adjacent_to_word(self):
        result = self.cli_parser.parse('aws ec2 stop- ')
        self.assert_parsed_results_equal(
            result,
            current_command='ec2',
            parsed_params={},
            global_params={},
            lineage=['aws'],
            current_fragment='',
            unparsed_items=['stop-'],
        )

    def test_current_fragment_populated_on_work_break(self):
        result = self.cli_parser.parse('aws ec2 ')
        self.assert_parsed_results_equal(
            result,
            current_command='ec2',
            parsed_params={},
            global_params={},
            lineage=['aws'],
            current_fragment='',
        )

    def test_current_fragment_can_be_option(self):
        result = self.cli_parser.parse('aws ec2 stop-instances --inst')
        # We should not have parsed the 'instance-ids'.
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            parsed_params={},
            lineage=['aws', 'ec2'],
            current_fragment='--inst',
        )

    def test_option_not_preserved_when_space_separated(self):
        result = self.cli_parser.parse('aws ec2 stop-instances --inst ')
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            parsed_params={},
            lineage=['aws', 'ec2'],
            current_fragment='',
            unparsed_items=['--inst'],
        )

    def test_can_have_unparsed_option_with_current_fragment(self):
        result = self.cli_parser.parse('aws ec2 stop-instances --inst foo')
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            parsed_params={},
            lineage=['aws', 'ec2'],
            current_fragment='foo',
            unparsed_items=['--inst'],
        )

    def test_unknown_option_does_not_consume_arg(self):
        # In this case we're unlikely to offer any helpful
        # auto-completion, but we still need to decided where
        # we should put the 'foo' value.  I think it makes the
        # most sense to put this under "unparsed_items".
        result = self.cli_parser.parse('aws ec2 stop-instances --inst foo ')
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            parsed_params={},
            lineage=['aws', 'ec2'],
            current_fragment='',
            unparsed_items=['--inst', 'foo'],
        )

    def test_can_handle_multiple_unknown_options(self):
        result = self.cli_parser.parse('aws ec2 stop-instances --inst --foo ')
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            parsed_params={},
            lineage=['aws', 'ec2'],
            current_fragment='',
            unparsed_items=['--inst', '--foo'],
        )

    def test_can_handle_unparsed_values(self):
        result = self.cli_parser.parse('aws ec stop-insta ')
        self.assert_parsed_results_equal(
            result,
            current_command='aws',
            current_fragment='',
            unparsed_items=['ec', 'stop-insta'],
        )

    def test_can_handle_error_cases_gracefully(self):
        # The --foo-arg has nargs of None so this is an invalid command.
        # It would normally generate a parser error.  We should make sure
        # we handle this gracefully.
        result = self.cli_parser.parse(
            'aws ec2 stop-instances --foo-arg a b --'
        )
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            lineage=['aws', 'ec2'],
            current_fragment=None,
            unparsed_items=['b', '--'],
        )

    def test_current_fragment_populated_for_param_value(self):
        result = self.cli_parser.parse(
            'aws ec2 stop-instances --instance-ids i-'
        )
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            current_param='instance-ids',
            parsed_params={'instance-ids': None},
            lineage=['aws', 'ec2'],
            current_fragment='i-',
            unparsed_items=[],
        )

    def test_params_populated_unless_last_value(self):
        result = self.cli_parser.parse(
            'aws ec2 stop-instances --instance-ids i-1 i-'
        )
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            current_param='instance-ids',
            # Here i-1 is in the parsed params because we've already moved
            # on from that word, however, 'i-' is not added to the parsed
            # params dict because we're still parsing that value.  It goes in
            # the ``current_fragment`` attribute.
            parsed_params={'instance-ids': ['i-1']},
            lineage=['aws', 'ec2'],
            current_fragment='i-',
            unparsed_items=[],
        )

    def test_params_stop_populate_when_dashes_entered(self):
        result = self.cli_parser.parse(
            'aws ec2 stop-instances --instance-ids i-1 --'
        )
        self.assert_parsed_results_equal(
            result,
            current_command='stop-instances',
            current_param=None,
            parsed_params={'instance-ids': ['i-1']},
            lineage=['aws', 'ec2'],
            current_fragment='--',
            unparsed_items=[],
        )

    def test_curent_fragment_works_for_global_params(self):
        result = self.cli_parser.parse(
            'aws --endpoint-url http',
        )
        self.assert_parsed_results_equal(
            result,
            current_command='aws',
            current_param='endpoint-url',
            # Here i-1 is in the parsed params because we've already moved
            # on from that word, however, 'i-' is not added to the parsed
            # params dict because we're still parsing that value.  It goes in
            # the ``current_fragment`` attribute.
            parsed_params={},
            lineage=[],
            current_fragment='http',
            unparsed_items=[],
        )

    def test_parses_positional(self):
        result = self.cli_parser.parse(
            'aws logs tail mygroup',
        )
        self.assert_parsed_results_equal(
            result,
            current_command='tail',
            current_param='group_name',
            parsed_params={},
            lineage=['aws', 'logs'],
            current_fragment='mygroup',
            unparsed_items=[],
        )

    def test_parses_positional_done(self):
        result = self.cli_parser.parse(
            'aws logs tail mygroup ',
        )
        self.assert_parsed_results_equal(
            result,
            current_command='tail',
            current_param=None,
            parsed_params={'group_name': 'mygroup'},
            lineage=['aws', 'logs'],
            current_fragment='',
            unparsed_items=[],
        )

    def test_parses_positional_parsed_only_once(self):
        result = self.cli_parser.parse(
            'aws logs tail mygroup random-content ',
        )
        self.assert_parsed_results_equal(
            result,
            current_command='tail',
            current_param=None,
            parsed_params={'group_name': 'mygroup'},
            lineage=['aws', 'logs'],
            current_fragment='',
            unparsed_items=['random-content'],
        )

    def test_parse_positional_with_option(self):
        result = self.cli_parser.parse(
            'aws logs tail mygroup --filter-pattern pat',
        )
        self.assert_parsed_results_equal(
            result,
            current_command='tail',
            current_param='filter-pattern',
            parsed_params={'filter-pattern': None, 'group_name': 'mygroup'},
            lineage=['aws', 'logs'],
            current_fragment='pat',
            unparsed_items=[],
        )

    def test_parse_positional_with_option_before(self):
        result = self.cli_parser.parse(
            'aws logs tail --filter-pattern pattern mygroup',
        )
        self.assert_parsed_results_equal(
            result,
            current_command='tail',
            current_param='group_name',
            parsed_params={
                'filter-pattern': 'pattern',
            },
            lineage=['aws', 'logs'],
            current_fragment='mygroup',
            unparsed_items=[],
        )

    def test_parse_outfile_option_in_the_beginning(self):
        result = self.cli_parser.parse(
            'aws s3api get-object filename --bucket myBucket --key foo ',
        )
        self.assert_parsed_s3api_result_correct(result)

    def test_parse_outfile_option_in_the_middle(self):
        result = self.cli_parser.parse(
            'aws s3api get-object --bucket myBucket filename --key foo ',
        )
        self.assert_parsed_s3api_result_correct(result)

    def test_parse_outfile_option_at_the_end(self):
        result = self.cli_parser.parse(
            'aws s3api get-object --bucket myBucket --key foo filename ',
        )
        self.assert_parsed_s3api_result_correct(result)


class TestParseState(unittest.TestCase):
    def test_can_set_initial_state(self):
        state = parser.ParseState()
        self.assertIsNone(state.current_command)
        self.assertEqual(state.lineage, [])
        self.assertEqual(state.full_lineage, [])

    def test_initial_starting_command_has_correct_lineage(self):
        state = parser.ParseState()
        state.current_command = 'aws'
        self.assertEqual(state.current_command, 'aws')
        self.assertEqual(state.lineage, [])
        self.assertEqual(state.full_lineage, ['aws'])

    def test_can_add_new_state(self):
        state = parser.ParseState()
        state.current_command = 'aws'
        state.current_command = 'ec2'
        self.assertEqual(state.current_command, 'ec2')
        self.assertEqual(state.lineage, ['aws'])
        self.assertEqual(state.full_lineage, ['aws', 'ec2'])

    def test_multiple_lineage_has_dotted_name(self):
        state = parser.ParseState()
        state.current_command = 'aws'
        state.current_command = 'ec2'
        state.current_command = 'describe-instances'
        self.assertEqual(state.current_command, 'describe-instances')
        self.assertEqual(state.lineage, ['aws', 'ec2'])
        self.assertEqual(
            state.full_lineage, ['aws', 'ec2', 'describe-instances']
        )



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

INSTANCE_GROUPS_WITH_AUTOSCALING_POLICY_ARG = (
    ' InstanceGroupType=MASTER,InstanceCount=1,InstanceType=d2.xlarge '
    'InstanceGroupType=TASK,InstanceType=d2.xlarge,InstanceCount=2,'
    'AutoScalingPolicy={Constraints={MinCapacity=1,MaxCapacity=2},'
    'Rules=[{Name=TestRule,Description=TestDescription,Action={Market=ON_DEMAND,'
    'SimpleScalingPolicyConfiguration={AdjustmentType=EXACT_CAPACITY,ScalingAdjustment=2,CoolDown=5}},'
    'Trigger={CloudWatchAlarmDefinition={ComparisonOperator=GREATER_THAN,'
    'EvaluationPeriods=5,MetricName=TestMetric,Namespace=EMR,Period=3,Statistic=MAXIMUM,'
    'Threshold=4.565,Unit=NONE,Dimensions=[{Key=TestKey,Value=TestValue}]}}}]}'
)

INSTANCE_GROUPS_WITH_EBS_VOLUME_ARG = (
    ' InstanceGroupType=MASTER,InstanceCount=1,InstanceType=d2.xlarge '
    'InstanceGroupType=CORE,InstanceType=d2.xlarge,InstanceCount=2,EbsConfiguration={EbsOptimized=true,EbsBlockDeviceConfigs=[{VolumeSpecification={VolumeType=gp2,SizeInGB=100,Iops=100},VolumesPerInstance=4},{VolumeSpecification={VolumeType=gp3,SizeInGB=100,Iops=3000,Throughput=200},VolumesPerInstance=1},{VolumeSpecification={VolumeType=gp3,SizeInGB=100}}]}'
)

INSTANCE_GROUPS_WITH_EBS_VOLUME_MISSING_VOLTYPE_ARG = (
    ' InstanceGroupType=MASTER,InstanceCount=1,InstanceType=d2.xlarge '
    'InstanceGroupType=CORE,InstanceType=d2.xlarge,InstanceCount=2,EbsConfiguration={EbsOptimized=true,EbsBlockDeviceConfigs=[{VolumeSpecification={SizeInGB=100,Iops=100},VolumesPerInstance=4},{VolumeSpecification={VolumeType=gp2,SizeInGB=100,Iops=100}}]}'
)

INSTANCE_GROUPS_WITH_EBS_VOLUME_MISSING_SIZE_ARG = (
    ' InstanceGroupType=MASTER,InstanceCount=1,InstanceType=d2.xlarge '
    'InstanceGroupType=CORE,InstanceType=d2.xlarge,InstanceCount=2,EbsConfiguration={EbsOptimized=true,EbsBlockDeviceConfigs=[{VolumeSpecification={VolumeType=gp2,Iops=100},VolumesPerInstance=4},{VolumeSpecification={VolumeType=gp2,SizeInGB=100,Iops=100}}]}'
)

INSTANCE_GROUPS_WITH_EBS_VOLUME_MISSING_VOLSPEC_ARG = (
    ' InstanceGroupType=MASTER,InstanceCount=1,InstanceType=d2.xlarge '
    'InstanceGroupType=CORE,InstanceType=d2.xlarge,InstanceCount=2,EbsConfiguration={EbsOptimized=true}'
)

INSTANCE_GROUPS_WITH_EBS_VOLUME_MISSING_IOPS_ARG = (
    ' InstanceGroupType=MASTER,InstanceCount=1,InstanceType=d2.xlarge '
    'InstanceGroupType=CORE,InstanceType=d2.xlarge,InstanceCount=2,EbsConfiguration={EbsOptimized=true,EbsBlockDeviceConfigs=[{VolumeSpecification={VolumeType=gp2,SizeInGB=100},VolumesPerInstance=4}]}'
)

MULTIPLE_INSTANCE_GROUPS_WITH_EBS_VOLUMES_VOLUME_ARG = (
    ' InstanceGroupType=MASTER,InstanceCount=1,InstanceType=d2.xlarge,EbsConfiguration={EbsOptimized=true,EbsBlockDeviceConfigs=[{VolumeSpecification={Iops=100,VolumeType=gp2,SizeInGB=20},VolumesPerInstance=4}]} '
    'InstanceGroupType=CORE,InstanceType=d2.xlarge,InstanceCount=2,EbsConfiguration={EbsOptimized=true,EbsBlockDeviceConfigs=[{VolumeSpecification={VolumeType=gp2,SizeInGB=100}}]}'
)

INSTANCE_GROUPS_WITH_EBS = [
    {
        'InstanceCount': 1,
        'InstanceRole': 'MASTER',
        'InstanceType': 'd2.xlarge',
        'Market': 'ON_DEMAND',
        'Name': 'MASTER',
    },
    {
        'EbsConfiguration': {
            'EbsOptimized': True,
            'EbsBlockDeviceConfigs': [
                {
                    'VolumeSpecification': {
                        'Iops': 100,
                        'SizeInGB': 100,
                        'VolumeType': 'gp2',
                    },
                    'VolumesPerInstance': 4,
                },
                {
                    'VolumeSpecification': {
                        'Iops': 3000,
                        'SizeInGB': 100,
                        'VolumeType': 'gp3',
                        'Throughput': 200,
                    },
                    'VolumesPerInstance': 1,
                },
                {
                    'VolumeSpecification': {
                        'SizeInGB': 100,
                        'VolumeType': 'gp3',
                    }
                },
            ],
        },
        'InstanceCount': 2,
        'InstanceRole': 'CORE',
        'InstanceType': 'd2.xlarge',
        'Market': 'ON_DEMAND',
        'Name': 'CORE',
    },
]

INSTANCE_GROUPS_WITH_EBS_VOLUME_MISSING_IOPS = [
    {
        'InstanceCount': 1,
        'InstanceRole': 'MASTER',
        'InstanceType': 'd2.xlarge',
        'Market': 'ON_DEMAND',
        'Name': 'MASTER',
    },
    {
        'EbsConfiguration': {
            'EbsOptimized': True,
            'EbsBlockDeviceConfigs': [
                {
                    'VolumeSpecification': {
                        'SizeInGB': 100,
                        'VolumeType': 'gp2',
                    },
                    'VolumesPerInstance': 4,
                }
            ],
        },
        'InstanceCount': 2,
        'InstanceRole': 'CORE',
        'InstanceType': 'd2.xlarge',
        'Market': 'ON_DEMAND',
        'Name': 'CORE',
    },
]

INSTANCE_GROUPS_WITH_EBS_VOLUME_MISSING_VOLSPEC = [
    {
        'InstanceCount': 1,
        'InstanceRole': 'MASTER',
        'InstanceType': 'd2.xlarge',
        'Market': 'ON_DEMAND',
        'Name': 'MASTER',
    },
    {
        'EbsConfiguration': {'EbsOptimized': True},
        'InstanceCount': 2,
        'InstanceRole': 'CORE',
        'InstanceType': 'd2.xlarge',
        'Market': 'ON_DEMAND',
        'Name': 'CORE',
    },
]

MULTIPLE_INSTANCE_GROUPS_WITH_EBS_VOLUMES = [
    {
        'EbsConfiguration': {
            'EbsOptimized': True,
            'EbsBlockDeviceConfigs': [
                {
                    'VolumeSpecification': {
                        'Iops': 100,
                        'SizeInGB': 20,
                        'VolumeType': 'gp2',
                    },
                    'VolumesPerInstance': 4,
                }
            ],
        },
        'InstanceCount': 1,
        'InstanceRole': 'MASTER',
        'InstanceType': 'd2.xlarge',
        'Market': 'ON_DEMAND',
        'Name': 'MASTER',
    },
    {
        'EbsConfiguration': {
            'EbsOptimized': True,
            'EbsBlockDeviceConfigs': [
                {'VolumeSpecification': {'SizeInGB': 100, 'VolumeType': 'gp2'}}
            ],
        },
        'InstanceCount': 2,
        'InstanceRole': 'CORE',
        'InstanceType': 'd2.xlarge',
        'Market': 'ON_DEMAND',
        'Name': 'CORE',
    },
]

INSTANCE_GROUPS_WITH_AUTOSCALING_POLICY = [
    {
        'InstanceCount': 1,
        'InstanceRole': 'MASTER',
        'InstanceType': 'd2.xlarge',
        'Market': 'ON_DEMAND',
        'Name': 'MASTER',
    },
    {
        'InstanceCount': 2,
        'InstanceRole': 'TASK',
        'InstanceType': 'd2.xlarge',
        'Market': 'ON_DEMAND',
        'Name': 'TASK',
        'AutoScalingPolicy': {
            'Constraints': {'MinCapacity': 1, 'MaxCapacity': 2},
            'Rules': [
                {
                    'Name': 'TestRule',
                    'Description': 'TestDescription',
                    'Action': {
                        'Market': 'ON_DEMAND',
                        'SimpleScalingPolicyConfiguration': {
                            'AdjustmentType': 'EXACT_CAPACITY',
                            'ScalingAdjustment': 2,
                            'CoolDown': 5,
                        },
                    },
                    'Trigger': {
                        'CloudWatchAlarmDefinition': {
                            'ComparisonOperator': 'GREATER_THAN',
                            'Dimensions': [
                                {'Key': 'TestKey', 'Value': 'TestValue'}
                            ],
                            'EvaluationPeriods': 5,
                            'MetricName': 'TestMetric',
                            'Namespace': 'EMR',
                            'Period': 3,
                            'Statistic': 'MAXIMUM',
                            'Threshold': 4.565,
                            'Unit': 'NONE',
                        }
                    },
                }
            ],
        },
    },
]



#!/usr/bin/env python
# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import awscli.clidriver
from awscli.testutils import BaseAWSCommandParamsTest


class TestSetQueueAttributes(BaseAWSCommandParamsTest):
    prefix = 'sqs set-queue-attributes'
    queue_url = 'https://queue.amazonaws.com/4444/testcli'

    def test_one(self):
        cmdline = self.prefix + ' --queue-url %s' % self.queue_url
        cmdline += ' --attributes {"VisibilityTimeout":"15"}'
        result = {
            'QueueUrl': self.queue_url,
            'Attributes': {'VisibilityTimeout': '15'},
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_shorthand(self):
        cmdline = self.prefix + ' --queue-url %s' % self.queue_url
        cmdline += ' --attributes VisibilityTimeout=15'
        result = {
            'QueueUrl': self.queue_url,
            'Attributes': {'VisibilityTimeout': '15'},
        }
        self.assert_params_for_cmd(cmdline, result)


if __name__ == "__main__":
    unittest.main()



import pytest

from awscli.customizations.ecs.expressgateway.color_utils import ColorUtils


@pytest.fixture
def color_utils():
    return ColorUtils()


def test_make_cyan(color_utils):
    result = color_utils.make_cyan("test text")
    # Should be cyan text with reset
    assert result == "\033[36mtest text\033[0m"


def test_make_cyan_no_color(color_utils):
    result = color_utils.make_cyan("test text", use_color=False)
    assert result == "test text"


def test_make_status_symbol_active(color_utils):
    result = color_utils.make_status_symbol("ACTIVE", "⠋")
    # Should return green checkmark with reset
    assert result == "\033[32m✓ \033[0m"


def test_make_status_symbol_active_no_color(color_utils):
    result = color_utils.make_status_symbol("ACTIVE", "⠋", use_color=False)
    assert result == "✓ "


def test_make_status_symbol_failed(color_utils):
    result = color_utils.make_status_symbol("FAILED", "⠋")
    # Should return red X with reset
    assert result == "\033[31mX \033[0m"


def test_make_status_symbol_failed_no_color(color_utils):
    result = color_utils.make_status_symbol("FAILED", "⠋", use_color=False)
    assert result == "X "


def test_make_status_symbol_provisioning(color_utils):
    result = color_utils.make_status_symbol("PROVISIONING", "⠋")
    # Should return purple spinner with reset
    assert result == "\033[35m⠋ \033[0m"


def test_make_status_symbol_none_status(color_utils):
    result = color_utils.make_status_symbol(None, "⠋")
    # Should return green checkmark with reset
    assert result == "\033[32m✓ \033[0m"


def test_color_by_status_active(color_utils):
    result = color_utils.color_by_status("test text", "ACTIVE")
    # Should be green text with reset
    assert result == "\033[32mtest text\033[0m"


def test_color_by_status_active_no_color(color_utils):
    result = color_utils.color_by_status(
        "test text", "ACTIVE", use_color=False
    )
    assert result == "test text"


def test_color_by_status_failed(color_utils):
    result = color_utils.color_by_status("test text", "FAILED")
    # Should be red text with reset
    assert result == "\033[31mtest text\033[0m"


def test_color_by_status_failed_no_color(color_utils):
    result = color_utils.color_by_status(
        "test text", "FAILED", use_color=False
    )
    assert result == "test text"


def test_color_by_status_provisioning(color_utils):
    result = color_utils.color_by_status("test text", "PROVISIONING")
    # Should be purple text with reset
    assert result == "\033[35mtest text\033[0m"


def test_color_by_status_none_status(color_utils):
    result = color_utils.color_by_status("test text", None)
    # Should return green text with reset
    assert result == "\033[32mtest text\033[0m"


def test_color_by_status_unknown_status(color_utils):
    result = color_utils.color_by_status("test text", "UNKNOWN_STATUS")
    # Should return purple text with reset
    assert result == "\033[35mtest text\033[0m"


def test_make_status_symbol_deleted(color_utils):
    result = color_utils.make_status_symbol("DELETED", "⠋")
    # Should return yellow dash with reset
    assert result == "\033[33m— \033[0m"


def test_color_by_status_deleted(color_utils):
    result = color_utils.color_by_status("test text", "DELETED")
    # Should be yellow text with reset
    assert result == "\033[33mtest text\033[0m"


def test_make_cyan_empty_string(color_utils):
    result = color_utils.make_cyan("")
    assert isinstance(result, str)


def test_color_by_status_empty_text(color_utils):
    result = color_utils.color_by_status("", "ACTIVE")
    assert isinstance(result, str)


def test_color_by_status_case_sensitive(color_utils):
    # Test that status matching is case sensitive
    result_upper = color_utils.color_by_status("test", "ACTIVE")
    result_lower = color_utils.color_by_status("test", "active")
    # ACTIVE should be green, active should be purple (case sensitive)
    assert result_upper == "\033[32mtest\033[0m"  # Green for ACTIVE
    assert result_lower == "\033[35mtest\033[0m"  # Purple for active


def test_make_status_symbol_case_sensitive(color_utils):
    # Test that status matching is case sensitive
    result_upper = color_utils.make_status_symbol("ACTIVE", "⠋")
    result_lower = color_utils.make_status_symbol("active", "⠋")
    # ACTIVE should be checkmark, active should be spinner (case sensitive)
    assert result_upper == "\033[32m✓ \033[0m"  # Checkmark for ACTIVE
    assert result_lower == "\033[35m⠋ \033[0m"  # Spinner for active



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.wizard import loader
from awscli.testutils import unittest


class TestLoader(unittest.TestCase):
    def setUp(self):
        self.loader = loader.WizardLoader()

    def test_can_list_commands_with_wizards(self):
        # This list is going to change over time so we'll only pick a few
        # commands as a sanity check.
        commands = self.loader.list_commands_with_wizards()
        self.assertIn('configure', commands)
        self.assertIn('iam', commands)

    def test_can_list_wizards_for_service(self):
        wizard_names = self.loader.list_available_wizards('iam')
        self.assertIn('new-role', wizard_names)

    def test_can_load_wizard_file(self):
        loaded = self.loader.load_wizard('iam', 'new-role')
        self.assertIn('version', loaded)

    def test_exception_raised_when_wizard_does_not_exists(self):
        with self.assertRaises(loader.WizardNotExistError):
            self.loader.load_wizard('iam', 'wizard-foo-does-not-exist')

    def test_wizard_exists(self):
        self.assertTrue(self.loader.wizard_exists('iam', 'new-role'))



# Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import contextlib
import glob
import json
import os
import re
import shlex
import shutil
import subprocess
import sys
import venv
from typing import Any, Callable, Dict, List, Optional

from constants import BOOTSTRAP_REQUIREMENTS, IS_WINDOWS, ROOT_DIR

PACKAGE_NAME = re.compile(r"(?P<name>[A-Za-z][A-Za-z0-9_\.\-]+)(?P<rest>.+)")
CONSTRAINT = re.compile(r"(?P<comparison>[=\<\>]+)(?P<version>.+)")
COMPARISONS: Dict[str, Callable[[List[int], List[int]], bool]] = {
    '==': lambda a, b: a == b,
    '>': lambda a, b: a > b,
    '>=': lambda a, b: a >= b,
    '<': lambda a, b: a < b,
    '<=': lambda a, b: a <= b,
}


class UnmetDependenciesException(Exception):
    def __init__(self, unmet_deps, in_venv, reason=None):
        pip_install_command_args = ["-m", "pip", "install", "--prefer-binary"]
        msg = "Environment requires following Python dependencies:\n\n"
        for package, actual_version, required in unmet_deps:
            msg += (
                f"{package} (required: {required.constraints}) "
                f"(version installed: {actual_version})\n"
            )
            pip_install_command_args.append(
                f'{package}{required.string_constraints()}'
            )

        if reason:
            msg += f"\n{reason}\n"

        msg += (
            "\n"
            "We recommend using --with-download-deps flag to automatically create a "
            "virtualenv and download the dependencies.\n\n"
            "If you want to manage the dependencies yourself instead, run the following "
            "pip command:\n"
        )
        msg += f"{sys.executable} {shlex.join(pip_install_command_args)}\n"

        if not in_venv:
            msg += (
                "\nWe noticed you are not in a virtualenv.\nIf not using --with-download-deps "
                "we highly recommend using a virtualenv to prevent dependencies "
                "from being installed into your global "
                "Python environment.\n"
            )
        super().__init__(msg)


@contextlib.contextmanager
def cd(dirname):
    original = os.getcwd()
    os.chdir(dirname)
    try:
        yield
    finally:
        os.chdir(original)


class Requirement:
    def __init__(self, name: str, *constraints):
        self.name = name
        self.constraints = constraints

    def is_in_range(self, version: str) -> bool:
        return self._meets_constraints(version)

    def _meets_constraints(self, version):
        return all(
            self._meets_constraint(version, constraint)
            for constraint in self.constraints
        )

    def _meets_constraint(self, version, constraint) -> bool:
        match = CONSTRAINT.match(constraint)
        if not match:
            raise RuntimeError(f"Unknown version specifier {constraint}")
        comparison, constraint_version = match.group('comparison', 'version')
        version, constraint_version = self._normalize(
            version, constraint_version
        )

        compare_fn = COMPARISONS.get(comparison)
        if not compare_fn:
            raise RuntimeError(f"Unknown version range specifier {comparison}")
        return compare_fn(version, constraint_version)

    def _normalize(self, v1: str, v2: str):
        v1_parts = [int(v) for v in v1.split(".")]
        v2_parts = [int(v) for v in v2.split(".")]
        while (pad := len(v1_parts) - len(v2_parts)) != 0:
            if pad > 0:
                v2_parts.append(0)
            if pad < 0:
                v1_parts.append(0)
        return v1_parts, v2_parts

    def __eq__(self, other):
        if other is None:
            return False
        return (
            self.name == other.name and self.constraints == other.constraints
        )

    def string_constraints(self):
        return ','.join(self.constraints)


class ParseError(Exception):
    pass


def parse_requirements(lines_list):
    lines = iter(lines_list)
    for line in lines:
        if ';' in line:
            raise ParseError('Parser does not support env markers')
        if line.startswith('#'):
            continue
        if ' #' in line:
            line = line[: line.find(' #')]
        if line.endswith('\\'):
            line = line[:-2].strip()
            try:
                line += next(lines)
            except StopIteration:
                return
        yield _parse_req_line(line)


def _parse_req_line(line: str):
    match = PACKAGE_NAME.search(line)
    if not match:
        raise RuntimeError(f"Unrecognized dependency {line}")

    name, rest = match.group('name', 'rest')
    return Requirement(name, *rest.split(','))


def get_install_requires():
    try:
        import flit_core.buildapi
    except ImportError:
        flit_core_exception = get_flit_core_unmet_exception()
        raise flit_core_exception

    with cd(ROOT_DIR):
        requires = flit_core.buildapi.get_requires_for_build_wheel()
    # Generation of the auto-complete index requires importing from the
    # awscli package and iterating over the commands from the clidriver. In
    # order to be able to do this, it requires all of the CLI's runtime
    # dependencies to be present to avoid import errors.
    dependency_block_re = re.compile(
        r"dependencies = \[([\s\S]+?)\]", re.MULTILINE
    )
    extract_dependencies_re = re.compile(r'"(.+)"')
    with open(ROOT_DIR / "pyproject.toml") as f:
        data = f.read()
    raw_dependencies = dependency_block_re.findall(data)[0]
    dependencies = extract_dependencies_re.findall(raw_dependencies)
    return dependencies


def get_flit_core_unmet_exception():
    in_venv = sys.prefix != sys.base_prefix
    with open(BOOTSTRAP_REQUIREMENTS) as f:
        flit_core_req = [l for l in f.read().split('\n') if 'flit_core' in l]
    return UnmetDependenciesException(
        [('flit_core', None, list(parse_requirements(flit_core_req))[0])],
        in_venv,
        reason=(
            'flit_core is needed ahead of time in order to parse the '
            'rest of the requirements.'
        ),
    )


class Utils:
    def isdir(self, path: str) -> bool:
        return os.path.isdir(path)

    def islink(self, path: str) -> bool:
        return os.path.islink(path)

    def remove(self, path: str):
        os.remove(path)

    def makedirs(self, path: str):
        os.makedirs(path)

    def symlink(self, src: str, dst: str):
        os.symlink(src, dst)

    def read_file_lines(self, path: str) -> List[str]:
        return open(path).readlines()

    def write_file(self, path: str, content: str):
        with open(path, "w") as f:
            f.write(content)

    def path_exists(self, path: str) -> bool:
        return os.path.exists(path)

    def rmtree(self, path: str) -> None:
        shutil.rmtree(path)

    def glob(self, pattern: str, root: str, recursive=True):
        with cd(root):
            return glob.glob(pattern, recursive=recursive)

    def run(self, args: List[str], **kwargs: Dict[str, Any]):
        return subprocess.run(args, **kwargs)

    def copy_file(self, src: str, dst: str):
        print("Copying file %s -> %s" % (src, dst))
        shutil.copy2(src, dst)

    def copy_directory_contents_into(self, src: str, dst: str):
        print("Copying contents of %s into %s" % (src, dst))
        shutil.copytree(src, dst, dirs_exist_ok=True)

    def copy_directory(self, src: str, dst: str):
        print("Copying %s -> %s" % (src, dst))
        shutil.copytree(src, dst)

    def update_metadata(self, dirname, **kwargs):
        print("Update metadata values %s" % kwargs)
        metadata_file = os.path.join(
            dirname, "awscli", "data", "metadata.json"
        )
        with open(metadata_file) as f:
            metadata = json.load(f)
        for key, value in kwargs.items():
            metadata[key] = value
        with open(metadata_file, "w") as f:
            json.dump(metadata, f)

    def create_venv(self, name: str, with_pip: bool = True):
        venv.create(name, with_pip=with_pip)

    def get_script_header(self, python_exe_path: str) -> str:
        if IS_WINDOWS:
            return (
                f'@echo off & "{python_exe_path}" -x "%~f0" %* & goto :eof\n'
            )
        return f"#!{python_exe_path}\n"



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.exceptions import ProfileNotFound


class FakeConfigStore:
    def set_config_provider(self, *args, **kwargs):
        pass


class FakeSession:
    def __init__(
        self,
        all_variables,
        profile_does_not_exist=False,
        config_file_vars=None,
        environment_vars=None,
        credentials=None,
        profile=None,
        available_profiles=None,
    ):
        self.variables = all_variables
        self.profile_does_not_exist = profile_does_not_exist
        self.config = {}
        self.full_config = {}
        if config_file_vars is None:
            config_file_vars = {}
        self.config_file_vars = config_file_vars
        if environment_vars is None:
            environment_vars = {}
        if available_profiles is None:
            available_profiles = ['default']
        self._available_profiles = available_profiles
        self.environment_vars = environment_vars
        self._credentials = credentials
        self.profile = profile
        self.user_agent_extra = ''

    @property
    def available_profiles(self):
        return self._available_profiles

    def get_credentials(self):
        return self._credentials

    def get_scoped_config(self):
        if self.profile_does_not_exist:
            raise ProfileNotFound(profile='foo')
        return self.config

    def get_component(self, component, *args, **kwargs):
        if component == 'config_store':
            return FakeConfigStore()

    def get_config_variable(self, name, methods=None):
        if name == 'credentials_file':
            # The credentials_file var doesn't require a
            # profile to exist.
            return '~/fake_credentials_filename'
        if self.profile_does_not_exist and not name == 'config_file':
            raise ProfileNotFound(profile='foo')
        if methods is not None:
            if 'env' in methods:
                return self.environment_vars.get(name)
            elif 'config' in methods:
                return self.config_file_vars.get(name)
        else:
            return self.variables.get(name)

    def emit(self, event_name, **kwargs):
        pass

    def emit_first_non_none_response(self, *args, **kwargs):
        pass

    def _build_profile_map(self):
        if self.full_config is None:
            return None
        return self.full_config['profiles']



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from unittest.mock import Mock, patch

import pytest

from awscli.customizations.ecs import inject_commands
from awscli.customizations.ecs.monitormutatinggatewayservice import (
    MUTATION_HANDLERS,
    MonitoringResourcesArgument,
    MonitorMutatingGatewayService,
    register_monitor_mutating_gateway_service,
)


@pytest.fixture
def mock_watcher_class():
    """Fixture that provides a mock watcher class."""
    watcher_class = Mock()
    watcher_class.is_monitoring_available.return_value = True
    return watcher_class


@pytest.fixture
def mock_session():
    """Fixture that provides a mock session."""
    return Mock()


@pytest.fixture
def handler(mock_watcher_class):
    """Fixture that provides a MonitorMutatingGatewayService handler."""
    return MonitorMutatingGatewayService(
        'create-gateway-service',
        'DEPLOYMENT',
        watcher_class=mock_watcher_class,
    )


class TestMonitoringResourcesArgument:
    def test_add_to_parser(self):
        parser = Mock()
        arg = MonitoringResourcesArgument('monitor-resources')
        arg.add_to_parser(parser)

        parser.add_argument.assert_called_once()


class TestMonitorMutatingGatewayService:
    def setup_method(self):
        self.mock_watcher_class = Mock()
        self.mock_watcher_class.is_monitoring_available.return_value = True
        self.handler = MonitorMutatingGatewayService(
            'create-gateway-service',
            'DEPLOYMENT',
            watcher_class=self.mock_watcher_class,
        )

    def test_init(self):
        assert self.handler.api == 'create-gateway-service'
        assert self.handler.default_resource_view == 'DEPLOYMENT'
        assert self.handler.api_pascal_case == 'CreateGatewayService'
        assert self.handler.session is None
        assert self.handler.parsed_globals is None

    def test_pascal_case_conversion(self):
        handler = MonitorMutatingGatewayService(
            'update-gateway-service', 'DEPLOYMENT'
        )
        assert handler.api_pascal_case == 'UpdateGatewayService'

        handler = MonitorMutatingGatewayService(
            'delete-gateway-service', 'RESOURCE'
        )
        assert handler.api_pascal_case == 'DeleteGatewayService'

    def test_before_building_argument_table_parser(self):
        session = Mock()

        self.handler.before_building_argument_table_parser(session)

        assert self.handler.session == session

    def test_building_argument_table(self):
        argument_table = {}
        session = Mock()

        self.handler.building_argument_table(argument_table, session)

        assert 'monitor-resources' in argument_table
        assert isinstance(
            argument_table['monitor-resources'], MonitoringResourcesArgument
        )

    def test_operation_args_parsed_with_monitor_resources_true(self):
        parsed_args = Mock()
        parsed_args.monitor_resources = True
        parsed_globals = Mock()

        self.handler.operation_args_parsed(parsed_args, parsed_globals)

        assert self.handler.effective_resource_view

    def test_operation_args_parsed_with_monitor_resources_false(self):
        parsed_args = Mock()
        parsed_args.monitor_resources = False
        parsed_globals = Mock()

        self.handler.operation_args_parsed(parsed_args, parsed_globals)

        assert not self.handler.effective_resource_view

    def test_after_call_with_monitoring_enabled(self):
        # Setup
        mock_watcher_class = Mock()
        mock_watcher = Mock()
        mock_watcher_class.return_value = mock_watcher

        handler = MonitorMutatingGatewayService(
            'create-express-gateway-service',
            'DEPLOYMENT',
            watcher_class=mock_watcher_class,
        )

        mock_session = Mock()
        mock_parsed_globals = Mock()
        mock_parsed_globals.region = 'us-west-2'
        mock_parsed_globals.endpoint_url = (
            'https://ecs.us-west-2.amazonaws.com'
        )
        mock_parsed_globals.verify_ssl = True

        mock_ecs_client = Mock()
        mock_session.create_client.return_value = mock_ecs_client

        handler.session = mock_session
        handler.parsed_globals = mock_parsed_globals
        handler.effective_resource_view = 'DEPLOYMENT'
        handler.effective_resource_view = 'DEPLOYMENT'

        parsed = {
            'service': {
                'serviceArn': 'arn:aws:ecs:us-west-2:123456789:service/test-service'
            }
        }
        context = {}
        http_response = Mock()
        http_response.status_code = 200

        # Execute
        handler.after_call(parsed, context, http_response)

        # Verify monitoring was initiated
        mock_watcher_class.assert_called_once()
        mock_watcher.exec.assert_called_once()

    def test_after_call_with_monitoring_disabled(self):
        # Setup
        mock_watcher_class = Mock()
        handler = MonitorMutatingGatewayService(
            'create-gateway-service',
            'DEPLOYMENT',
            watcher_class=mock_watcher_class,
        )

        mock_session = Mock()
        mock_parsed_globals = Mock()
        handler.session = mock_session
        handler.parsed_globals = mock_parsed_globals
        handler.effective_resource_view = None

        parsed = {
            'service': {
                'serviceArn': 'arn:aws:ecs:us-west-2:123456789:service/test-service'
            }
        }
        context = {}
        http_response = Mock()
        http_response.status_code = 200

        # Execute
        handler.after_call(parsed, context, http_response)

        # Verify monitoring was skipped
        mock_watcher_class.assert_not_called()

    def test_after_call_no_session(self):
        # Setup
        mock_watcher_class = Mock()
        handler = MonitorMutatingGatewayService(
            'create-gateway-service',
            'DEPLOYMENT',
            watcher_class=mock_watcher_class,
        )

        handler.session = None
        handler.parsed_globals = None
        handler.effective_resource_view = 'DEPLOYMENT'

        parsed = {
            'service': {
                'serviceArn': 'arn:aws:ecs:us-west-2:123456789:service/test-service'
            }
        }
        context = {}
        http_response = Mock()
        http_response.status_code = 200

        # Execute
        handler.after_call(parsed, context, http_response)

        # Verify monitoring was skipped due to missing session
        mock_watcher_class.assert_not_called()

    def test_events(self):
        events = self.handler.events()

        expected_events = [
            "before-building-argument-table-parser.ecs.create-gateway-service",
            "building-argument-table.ecs.create-gateway-service",
            "operation-args-parsed.ecs.create-gateway-service",
            "after-call.ecs.CreateGatewayService",
        ]

        assert len(events) == 4
        for i, (event_name, handler_method) in enumerate(events):
            assert event_name == expected_events[i]
            assert callable(handler_method)


class TestMutationHandlers:
    def test_mutation_handlers_configuration(self):
        assert len(MUTATION_HANDLERS) == 3

        # Test create handler
        create_handler = MUTATION_HANDLERS[0]
        assert create_handler.api == 'create-express-gateway-service'
        assert create_handler.default_resource_view == 'DEPLOYMENT'

        # Test update handler
        update_handler = MUTATION_HANDLERS[1]
        assert update_handler.api == 'update-express-gateway-service'
        assert update_handler.default_resource_view == 'DEPLOYMENT'

        # Test delete handler
        delete_handler = MUTATION_HANDLERS[2]
        assert delete_handler.api == 'delete-express-gateway-service'
        assert delete_handler.default_resource_view == 'RESOURCE'


class TestRegisterFunction:
    def test_register_monitor_mutating_gateway_service(self):
        mock_event_handler = Mock()

        register_monitor_mutating_gateway_service(mock_event_handler)

        # Should register 4 events per handler * 3 handlers = 12 total registrations
        assert mock_event_handler.register.call_count == 12

        # Verify some specific event registrations
        registered_events = [
            call[0][0] for call in mock_event_handler.register.call_args_list
        ]

        # Check that create-express-gateway-service events are registered
        assert (
            'before-building-argument-table-parser.ecs.create-express-gateway-service'
            in registered_events
        )
        assert (
            'after-call.ecs.CreateExpressGatewayService' in registered_events
        )


class TestMonitorModeParameter:
    """Tests for --monitor-mode parameter functionality."""

    def test_monitor_mode_argument_added_to_table(self, handler):
        """Test that --monitor-mode is added to argument table."""
        argument_table = {}
        session = Mock()

        handler.building_argument_table(argument_table, session)

        assert 'monitor-mode' in argument_table

    @patch('sys.stdout.isatty', return_value=True)
    def test_operation_args_parsed_with_monitor_mode_and_resources(
        self, mock_isatty, handler, mock_session
    ):
        """Test operation_args_parsed with both --monitor-mode and --monitor-resources."""
        handler.session = mock_session
        parsed_args = Mock()
        parsed_args.monitor_resources = 'RESOURCE'
        parsed_args.monitor_mode = 'TEXT-ONLY'
        parsed_globals = Mock()

        # Should not raise
        handler.operation_args_parsed(parsed_args, parsed_globals)

        assert handler.effective_resource_view == 'RESOURCE'
        assert handler.effective_mode == 'TEXT-ONLY'

    @patch('sys.stdout.isatty', return_value=True)
    def test_operation_args_parsed_with_monitor_mode_without_resources_raises(
        self, mock_isatty, handler, mock_session
    ):
        """Test operation_args_parsed with --monitor-mode but no --monitor-resources raises ValueError."""
        handler.session = mock_session
        parsed_args = Mock()
        parsed_args.monitor_resources = None
        parsed_args.monitor_mode = 'TEXT-ONLY'
        parsed_globals = Mock()

        with pytest.raises(ValueError) as exc_info:
            handler.operation_args_parsed(parsed_args, parsed_globals)

        assert (
            '--monitor-mode can only be used with --monitor-resources'
            in str(exc_info.value)
        )

    @patch('sys.stdout.isatty', return_value=True)
    def test_operation_args_parsed_defaults_mode_to_interactive(
        self, mock_isatty, handler, mock_session
    ):
        """Test operation_args_parsed defaults mode to INTERACTIVE when not specified."""
        handler.session = mock_session
        parsed_args = Mock()
        parsed_args.monitor_resources = 'DEPLOYMENT'
        parsed_args.monitor_mode = None
        parsed_globals = Mock()

        handler.operation_args_parsed(parsed_args, parsed_globals)

        assert handler.effective_mode == 'INTERACTIVE'

    @patch('sys.stdout.isatty', return_value=True)
    def test_operation_args_parsed_without_monitor_resources(
        self, mock_isatty, handler, mock_session
    ):
        """Test operation_args_parsed disables monitoring when --monitor-resources not provided."""
        handler.session = mock_session
        parsed_args = Mock()
        parsed_args.monitor_resources = None
        parsed_args.monitor_mode = None
        parsed_globals = Mock()

        handler.operation_args_parsed(parsed_args, parsed_globals)

        assert handler.effective_resource_view is None

    @patch('sys.stdout.isatty', return_value=True)
    def test_after_call_with_interactive_mode(
        self, mock_isatty, mock_session, mock_watcher_class
    ):
        """Test monitoring starts with interactive mode when specified."""
        handler = MonitorMutatingGatewayService(
            'create-express-gateway-service',
            'DEPLOYMENT',
            watcher_class=mock_watcher_class,
        )

        mock_watcher = Mock()
        mock_watcher_class.return_value = mock_watcher

        mock_parsed_globals = Mock()
        mock_parsed_globals.region = 'us-west-2'
        mock_parsed_globals.endpoint_url = None
        mock_parsed_globals.verify_ssl = True
        mock_parsed_globals.color = 'auto'

        mock_ecs_client = Mock()
        mock_session.create_client.return_value = mock_ecs_client

        handler.session = mock_session
        handler.parsed_globals = mock_parsed_globals
        handler.effective_resource_view = 'DEPLOYMENT'
        handler.effective_mode = 'INTERACTIVE'

        parsed = {
            'service': {
                'serviceArn': 'arn:aws:ecs:us-west-2:123456789:service/test-service'
            }
        }
        context = {}
        http_response = Mock()
        http_response.status_code = 200

        handler.after_call(parsed, context, http_response)

        # Verify watcher was created with correct parameters
        call_args = mock_watcher_class.call_args
        assert call_args is not None

        # Check positional arguments
        assert (
            call_args[0][1]
            == 'arn:aws:ecs:us-west-2:123456789:service/test-service'
        )  # service_arn
        assert call_args[0][2] == 'DEPLOYMENT'  # resource_view
        assert call_args[0][3] == 'INTERACTIVE'

    @patch('sys.stdout.isatty', return_value=False)
    def test_after_call_with_text_only_mode(
        self, mock_isatty, mock_session, mock_watcher_class
    ):
        """Test monitoring starts with text-only mode when specified."""
        handler = MonitorMutatingGatewayService(
            'create-express-gateway-service',
            'DEPLOYMENT',
            watcher_class=mock_watcher_class,
        )

        mock_watcher = Mock()
        mock_watcher_class.return_value = mock_watcher

        mock_parsed_globals = Mock()
        mock_parsed_globals.region = 'us-west-2'
        mock_parsed_globals.endpoint_url = None
        mock_parsed_globals.verify_ssl = True
        mock_parsed_globals.color = 'off'

        mock_ecs_client = Mock()
        mock_session.create_client.return_value = mock_ecs_client

        handler.session = mock_session
        handler.parsed_globals = mock_parsed_globals
        handler.effective_resource_view = 'RESOURCE'
        handler.effective_mode = 'TEXT-ONLY'

        parsed = {
            'service': {
                'serviceArn': 'arn:aws:ecs:us-west-2:123456789:service/test-service'
            }
        }
        context = {}
        http_response = Mock()
        http_response.status_code = 200

        handler.after_call(parsed, context, http_response)

        # Verify watcher was created with correct parameters
        call_args = mock_watcher_class.call_args
        assert call_args is not None

        # Check positional arguments
        assert (
            call_args[0][1]
            == 'arn:aws:ecs:us-west-2:123456789:service/test-service'
        )  # service_arn
        assert call_args[0][2] == 'RESOURCE'  # resource_view
        assert call_args[0][3] == 'TEXT-ONLY'

    @patch('sys.stdout.isatty', return_value=True)
    def test_after_call_with_color_on(
        self, mock_isatty, mock_session, mock_watcher_class
    ):
        """Test use_color=True when color='on'."""
        handler = MonitorMutatingGatewayService(
            'create-express-gateway-service',
            'DEPLOYMENT',
            watcher_class=mock_watcher_class,
        )

        mock_watcher = Mock()
        mock_watcher_class.return_value = mock_watcher

        mock_parsed_globals = Mock()
        mock_parsed_globals.region = 'us-west-2'
        mock_parsed_globals.endpoint_url = None
        mock_parsed_globals.verify_ssl = True
        mock_parsed_globals.color = 'on'

        mock_ecs_client = Mock()
        mock_session.create_client.return_value = mock_ecs_client

        handler.session = mock_session
        handler.parsed_globals = mock_parsed_globals
        handler.effective_resource_view = 'DEPLOYMENT'
        handler.effective_mode = 'INTERACTIVE'

        parsed = {
            'service': {
                'serviceArn': 'arn:aws:ecs:us-west-2:123456789:service/test-service'
            }
        }
        context = {}
        http_response = Mock()
        http_response.status_code = 200

        handler.after_call(parsed, context, http_response)

        # Check keyword arguments
        call_args = mock_watcher_class.call_args
        assert call_args[1]['use_color'] is True

    @patch('sys.stdout.isatty', return_value=False)
    def test_after_call_with_color_off(
        self, mock_isatty, mock_session, mock_watcher_class
    ):
        """Test use_color=False when color='off'."""
        handler = MonitorMutatingGatewayService(
            'create-express-gateway-service',
            'DEPLOYMENT',
            watcher_class=mock_watcher_class,
        )

        mock_watcher = Mock()
        mock_watcher_class.return_value = mock_watcher

        mock_parsed_globals = Mock()
        mock_parsed_globals.region = 'us-west-2'
        mock_parsed_globals.endpoint_url = None
        mock_parsed_globals.verify_ssl = True
        mock_parsed_globals.color = 'off'

        mock_ecs_client = Mock()
        mock_session.create_client.return_value = mock_ecs_client

        handler.session = mock_session
        handler.parsed_globals = mock_parsed_globals
        handler.effective_resource_view = 'DEPLOYMENT'
        handler.effective_mode = 'TEXT-ONLY'

        parsed = {
            'service': {
                'serviceArn': 'arn:aws:ecs:us-west-2:123456789:service/test-service'
            }
        }
        context = {}
        http_response = Mock()
        http_response.status_code = 200

        handler.after_call(parsed, context, http_response)

        call_args = mock_watcher_class.call_args
        assert call_args[1]['use_color'] is False

    @patch('sys.stdout.isatty', return_value=True)
    def test_after_call_with_color_auto_with_tty(
        self, mock_isatty, mock_session, mock_watcher_class
    ):
        """Test use_color=True when color='auto' with TTY."""
        handler = MonitorMutatingGatewayService(
            'create-express-gateway-service',
            'DEPLOYMENT',
            watcher_class=mock_watcher_class,
        )

        mock_watcher = Mock()
        mock_watcher_class.return_value = mock_watcher

        mock_parsed_globals = Mock()
        mock_parsed_globals.region = 'us-west-2'
        mock_parsed_globals.endpoint_url = None
        mock_parsed_globals.verify_ssl = True
        mock_parsed_globals.color = 'auto'

        mock_ecs_client = Mock()
        mock_session.create_client.return_value = mock_ecs_client

        handler.session = mock_session
        handler.parsed_globals = mock_parsed_globals
        handler.effective_resource_view = 'DEPLOYMENT'
        handler.effective_mode = 'INTERACTIVE'

        parsed = {
            'service': {
                'serviceArn': 'arn:aws:ecs:us-west-2:123456789:service/test-service'
            }
        }
        context = {}
        http_response = Mock()
        http_response.status_code = 200

        handler.after_call(parsed, context, http_response)

        call_args = mock_watcher_class.call_args
        assert call_args[1]['use_color'] is True

    @patch('sys.stdout.isatty', return_value=False)
    def test_after_call_with_color_auto_without_tty(
        self, mock_isatty, mock_session, mock_watcher_class
    ):
        """Test use_color=False when color='auto' without TTY."""
        handler = MonitorMutatingGatewayService(
            'create-express-gateway-service',
            'DEPLOYMENT',
            watcher_class=mock_watcher_class,
        )

        mock_watcher = Mock()
        mock_watcher_class.return_value = mock_watcher

        mock_parsed_globals = Mock()
        mock_parsed_globals.region = 'us-west-2'
        mock_parsed_globals.endpoint_url = None
        mock_parsed_globals.verify_ssl = True
        mock_parsed_globals.color = 'auto'

        mock_ecs_client = Mock()
        mock_session.create_client.return_value = mock_ecs_client

        handler.session = mock_session
        handler.parsed_globals = mock_parsed_globals
        handler.effective_resource_view = 'DEPLOYMENT'
        handler.effective_mode = 'TEXT-ONLY'

        parsed = {
            'service': {
                'serviceArn': 'arn:aws:ecs:us-west-2:123456789:service/test-service'
            }
        }
        context = {}
        http_response = Mock()
        http_response.status_code = 200

        handler.after_call(parsed, context, http_response)

        call_args = mock_watcher_class.call_args
        assert call_args[1]['use_color'] is False



import botocore.session
from botocore.stub import Stubber

from awscli.customizations.cloudformation import exceptions
from awscli.customizations.cloudformation.deployer import (
    ChangeSetResult,
    Deployer,
)
from awscli.testutils import mock
from tests.unit.customizations.cloudformation import BaseYAMLTest


class TestDeployer(BaseYAMLTest):
    def setUp(self):
        super(TestDeployer, self).setUp()
        client = botocore.session.get_session().create_client(
            'cloudformation', region_name="us-east-1"
        )
        self.stub_client = Stubber(client)

        self.changeset_prefix = "some-changeset-prefix"
        self.deployer = Deployer(client, self.changeset_prefix)

    def test_has_stack_success(self):
        stack_name = "stack_name"

        expected_params = {"StackName": stack_name}

        response = {"Stacks": [make_stack_obj(stack_name)]}

        self.stub_client.add_response(
            'describe_stacks', response, expected_params
        )

        with self.stub_client:
            response = self.deployer.has_stack(stack_name)
            self.assertTrue(response)

    def test_has_stack_no_stack(self):
        stack_name = "stack_name"
        expected_params = {"StackName": stack_name}

        # Response contains NO stack
        no_stack_response = {"Stacks": []}
        self.stub_client.add_response(
            'describe_stacks', no_stack_response, expected_params
        )
        with self.stub_client:
            response = self.deployer.has_stack(stack_name)
            self.assertFalse(response)

        # Response is a ClientError with a message that the stack does not exist
        self.stub_client.add_client_error(
            'describe_stacks',
            "ClientError",
            f"Stack with id {stack_name} does not exist",
        )
        with self.stub_client:
            response = self.deployer.has_stack(stack_name)
            self.assertFalse(response)

    def test_has_stack_review_in_progress(self):
        stack_name = "stack_name"
        expected_params = {"StackName": stack_name}

        # Response contains NO stack
        review_in_progress_response = {
            "Stacks": [make_stack_obj(stack_name, "REVIEW_IN_PROGRESS")]
        }
        self.stub_client.add_response(
            'describe_stacks', review_in_progress_response, expected_params
        )
        with self.stub_client:
            response = self.deployer.has_stack(stack_name)
            self.assertFalse(response)

    def test_has_stack_exception(self):
        self.stub_client.add_client_error(
            'describe_stacks', "ValidationError", "Service is bad"
        )
        with self.stub_client:
            with self.assertRaises(botocore.exceptions.ClientError):
                self.deployer.has_stack("stack_name")

    def test_create_changeset_success(self):
        stack_name = "stack_name"
        template = "template"
        parameters = [
            {"ParameterKey": "Key1", "ParameterValue": "Value"},
            {"ParameterKey": "Key2", "UsePreviousValue": True},
            {"ParameterKey": "Key3", "UsePreviousValue": False},
        ]
        # Parameters that Use Previous Value will be removed on stack creation
        # to either force CloudFormation to use the Default value, or ask user to specify a parameter
        filtered_parameters = [
            {"ParameterKey": "Key1", "ParameterValue": "Value"},
            {"ParameterKey": "Key3", "UsePreviousValue": False},
        ]
        capabilities = ["capabilities"]
        role_arn = "arn:aws:iam::1234567890:role"
        notification_arns = ["arn:aws:sns:region:1234567890:notify"]
        s3_uploader = None

        tags = [{"Key": "key1", "Value": "val1"}]

        # Case 1: Stack DOES NOT exist
        self.deployer.has_stack = mock.Mock()
        self.deployer.has_stack.return_value = False

        expected_params = {
            "ChangeSetName": botocore.stub.ANY,
            "StackName": stack_name,
            "TemplateBody": template,
            "ChangeSetType": "CREATE",
            "Parameters": filtered_parameters,
            "Capabilities": capabilities,
            "Description": botocore.stub.ANY,
            "RoleARN": role_arn,
            "NotificationARNs": notification_arns,
            "Tags": tags,
        }

        response = {"Id": "changeset ID"}

        self.stub_client.add_response(
            "create_change_set", response, expected_params
        )
        with self.stub_client:
            result = self.deployer.create_changeset(
                stack_name,
                template,
                parameters,
                capabilities,
                role_arn,
                notification_arns,
                s3_uploader,
                tags,
            )
            self.assertEqual(response["Id"], result.changeset_id)
            self.assertEqual("CREATE", result.changeset_type)

        # Case 2: Stack exists. We are updating it
        self.deployer.has_stack.return_value = True
        self.stub_client.add_response(
            "get_template_summary",
            {
                "Parameters": [
                    {"ParameterKey": parameter["ParameterKey"]}
                    for parameter in parameters
                ]
            },
            {"StackName": stack_name},
        )
        expected_params["ChangeSetType"] = "UPDATE"
        expected_params["Parameters"] = parameters
        self.stub_client.add_response(
            "create_change_set", response, expected_params
        )
        # template has new parameter but should not be included in
        # expected_params as no previous value
        parameters = list(parameters) + [
            {"ParameterKey": "New", "UsePreviousValue": True}
        ]
        with self.stub_client:
            result = self.deployer.create_changeset(
                stack_name,
                template,
                parameters,
                capabilities,
                role_arn,
                notification_arns,
                s3_uploader,
                tags,
            )
            self.assertEqual(response["Id"], result.changeset_id)
            self.assertEqual("UPDATE", result.changeset_type)

    def test_create_changeset_success_s3_bucket(self):
        stack_name = "stack_name"
        template = "template"
        template_url = "https://s3.amazonaws.com/bucket/file"
        parameters = [
            {"ParameterKey": "Key1", "ParameterValue": "Value"},
            {"ParameterKey": "Key2", "UsePreviousValue": True},
            {"ParameterKey": "Key3", "UsePreviousValue": False},
        ]
        # Parameters that Use Previous Value will be removed on stack creation
        # to either force CloudFormation to use the Default value, or ask user to specify a parameter
        filtered_parameters = [
            {"ParameterKey": "Key1", "ParameterValue": "Value"},
            {"ParameterKey": "Key3", "UsePreviousValue": False},
        ]
        capabilities = ["capabilities"]
        role_arn = "arn:aws:iam::1234567890:role"
        notification_arns = ["arn:aws:sns:region:1234567890:notify"]

        s3_uploader = mock.Mock()

        def to_path_style_s3_url(some_string, Version=None):
            return "https://s3.amazonaws.com/bucket/file"

        s3_uploader.to_path_style_s3_url = to_path_style_s3_url

        def upload_with_dedup(filename, extension):
            return "s3://bucket/file"

        s3_uploader.upload_with_dedup = upload_with_dedup

        # Case 1: Stack DOES NOT exist
        self.deployer.has_stack = mock.Mock()
        self.deployer.has_stack.return_value = False

        expected_params = {
            "ChangeSetName": botocore.stub.ANY,
            "StackName": stack_name,
            "TemplateURL": template_url,
            "ChangeSetType": "CREATE",
            "Parameters": filtered_parameters,
            "Capabilities": capabilities,
            "Description": botocore.stub.ANY,
            "RoleARN": role_arn,
            "Tags": [],
            "NotificationARNs": notification_arns,
        }

        response = {"Id": "changeset ID"}

        self.stub_client.add_response(
            "create_change_set", response, expected_params
        )
        with self.stub_client:
            result = self.deployer.create_changeset(
                stack_name,
                template,
                parameters,
                capabilities,
                role_arn,
                notification_arns,
                s3_uploader,
                [],
            )
            self.assertEqual(response["Id"], result.changeset_id)
            self.assertEqual("CREATE", result.changeset_type)

        # Case 2: Stack exists. We are updating it
        self.deployer.has_stack.return_value = True
        self.stub_client.add_response(
            "get_template_summary",
            {
                "Parameters": [
                    {"ParameterKey": parameter["ParameterKey"]}
                    for parameter in parameters
                ]
            },
            {"StackName": stack_name},
        )
        expected_params["ChangeSetType"] = "UPDATE"
        expected_params["Parameters"] = parameters
        # template has new parameter but should not be included in
        # expected_params as no previous value
        parameters = list(parameters) + [
            {"ParameterKey": "New", "UsePreviousValue": True}
        ]
        self.stub_client.add_response(
            "create_change_set", response, expected_params
        )
        with self.stub_client:
            result = self.deployer.create_changeset(
                stack_name,
                template,
                parameters,
                capabilities,
                role_arn,
                notification_arns,
                s3_uploader,
                [],
            )
            self.assertEqual(response["Id"], result.changeset_id)
            self.assertEqual("UPDATE", result.changeset_type)

    def test_create_changeset_exception(self):
        stack_name = "stack_name"
        template = "template"
        parameters = [
            {
                "ParameterKey": "Key1",
                "ParameterValue": "Value",
                "UsePreviousValue": True,
            }
        ]
        capabilities = ["capabilities"]
        role_arn = "arn:aws:iam::1234567890:role"
        notification_arns = ["arn:aws:sns:region:1234567890:notify"]
        s3_uploader = None
        tags = [{"Key": "key1", "Value": "val1"}]

        self.deployer.has_stack = mock.Mock()
        self.deployer.has_stack.return_value = False

        self.stub_client.add_client_error(
                'create_change_set', "Something is wrong", "Service is bad")
        with self.stub_client:
            with self.assertRaises(botocore.exceptions.ClientError):
                self.deployer.create_changeset(
                    stack_name,
                    template,
                    parameters,
                    capabilities,
                    role_arn,
                    notification_arns,
                    None,
                    tags,
                )

    def test_execute_changeset(self):
        stack_name = "stack_name"
        changeset_id = "changeset_id"

        expected_params = {
            "ChangeSetName": changeset_id,
            "StackName": stack_name,
            "DisableRollback": False,
        }

        self.stub_client.add_response(
            "execute_change_set", {}, expected_params
        )
        with self.stub_client:
            self.deployer.execute_changeset(changeset_id, stack_name)

    def test_execute_changeset_disable_rollback(self):
        stack_name = "stack_name"
        changeset_id = "changeset_id"
        disable_rollback = True

        expected_params = {
            "ChangeSetName": changeset_id,
            "StackName": stack_name,
            "DisableRollback": disable_rollback,
        }

        self.stub_client.add_response(
            "execute_change_set", {}, expected_params
        )
        with self.stub_client:
            self.deployer.execute_changeset(
                changeset_id, stack_name, disable_rollback
            )

    def test_execute_changeset_exception(self):
        stack_name = "stack_name"
        changeset_id = "changeset_id"

        self.stub_client.add_client_error(
                'execute_change_set', "Something is wrong", "Service is bad")
        with self.stub_client:
            with self.assertRaises(botocore.exceptions.ClientError):
                self.deployer.execute_changeset(changeset_id, stack_name)

    def test_create_and_wait_for_changeset_successful(self):
        stack_name = "stack_name"
        template = "template"
        parameters = [
            {
                "ParameterKey": "Key1",
                "ParameterValue": "Value",
                "UsePreviousValue": True,
            }
        ]
        capabilities = ["capabilities"]
        changeset_id = "changeset id"
        changeset_type = "changeset type"
        role_arn = "arn:aws:iam::1234567890:role"
        notification_arns = ["arn:aws:sns:region:1234567890:notify"]
        s3_uploader = None
        tags = [{"Key": "key1", "Value": "val1"}]

        self.deployer.create_changeset = mock.Mock()
        self.deployer.create_changeset.return_value = ChangeSetResult(
            changeset_id, changeset_type
        )

        self.deployer.wait_for_changeset = mock.Mock()

        result = self.deployer.create_and_wait_for_changeset(
            stack_name,
            template,
            parameters,
            capabilities,
            role_arn,
            notification_arns,
            s3_uploader,
            tags,
        )
        self.assertEqual(result.changeset_id, changeset_id)
        self.assertEqual(result.changeset_type, changeset_type)

    def test_create_and_wait_for_changeset_error_waiting_for_changeset(self):
        stack_name = "stack_name"
        template = "template"
        parameters = [
            {
                "ParameterKey": "Key1",
                "ParameterValue": "Value",
                "UsePreviousValue": True,
            }
        ]
        capabilities = ["capabilities"]
        changeset_id = "changeset id"
        changeset_type = "changeset type"
        role_arn = "arn:aws:iam::1234567890:role"
        notification_arns = ["arn:aws:sns:region:1234567890:notify"]
        s3_uploader = None
        tags = [{"Key": "key1", "Value": "val1"}]

        self.deployer.create_changeset = mock.Mock()
        self.deployer.create_changeset.return_value = ChangeSetResult(
            changeset_id, changeset_type
        )

        self.deployer.wait_for_changeset = mock.Mock()
        self.deployer.wait_for_changeset.side_effect = RuntimeError

        with self.assertRaises(RuntimeError):
            result = self.deployer.create_and_wait_for_changeset(
                stack_name,
                template,
                parameters,
                capabilities,
                role_arn,
                notification_arns,
                s3_uploader,
                tags,
            )

    def test_wait_for_changeset_no_changes(self):
        stack_name = "stack_name"
        changeset_id = "changeset-id"

        mock_client = mock.Mock()
        mock_deployer = Deployer(mock_client)
        mock_waiter = mock.Mock()
        mock_client.get_waiter.return_value = mock_waiter

        response = {
            "Status": "FAILED",
            "StatusReason": "The submitted information didn't contain changes.",
        }

        waiter_error = botocore.exceptions.WaiterError(
            name="name", reason="reason", last_response=response
        )
        mock_waiter.wait.side_effect = waiter_error

        with self.assertRaises(exceptions.ChangeEmptyError):
            mock_deployer.wait_for_changeset(changeset_id, stack_name)

        waiter_config = {'Delay': 5}
        mock_waiter.wait.assert_called_once_with(
            ChangeSetName=changeset_id,
            StackName=stack_name,
            WaiterConfig=waiter_config,
        )

        mock_client.get_waiter.assert_called_once_with(
            "change_set_create_complete"
        )

    def test_wait_for_changeset_no_changes_with_another_error_msg(self):
        stack_name = "stack_name"
        changeset_id = "changeset-id"

        mock_client = mock.Mock()
        mock_deployer = Deployer(mock_client)
        mock_waiter = mock.Mock()
        mock_client.get_waiter.return_value = mock_waiter

        response = {
            "Status": "FAILED",
            "StatusReason": "No updates are to be performed",
        }

        waiter_error = botocore.exceptions.WaiterError(
            name="name", reason="reason", last_response=response
        )
        mock_waiter.wait.side_effect = waiter_error

        with self.assertRaises(exceptions.ChangeEmptyError):
            mock_deployer.wait_for_changeset(changeset_id, stack_name)

        waiter_config = {'Delay': 5}
        mock_waiter.wait.assert_called_once_with(
            ChangeSetName=changeset_id,
            StackName=stack_name,
            WaiterConfig=waiter_config,
        )

        mock_client.get_waiter.assert_called_once_with(
            "change_set_create_complete"
        )

    def test_wait_for_changeset_failed_to_create_changeset(self):
        stack_name = "stack_name"
        changeset_id = "changeset-id"

        mock_client = mock.Mock()
        mock_deployer = Deployer(mock_client)
        mock_waiter = mock.Mock()
        mock_client.get_waiter.return_value = mock_waiter

        response = {"Status": "FAILED", "StatusReason": "some reason"}

        waiter_error = botocore.exceptions.WaiterError(
            name="name", reason="reason", last_response=response
        )
        mock_waiter.wait.side_effect = waiter_error

        with self.assertRaises(RuntimeError):
            mock_deployer.wait_for_changeset(changeset_id, stack_name)

        waiter_config = {'Delay': 5}
        mock_waiter.wait.assert_called_once_with(
            ChangeSetName=changeset_id,
            StackName=stack_name,
            WaiterConfig=waiter_config,
        )

        mock_client.get_waiter.assert_called_once_with(
            "change_set_create_complete"
        )

    def test_wait_for_execute_no_changes(self):
        stack_name = "stack_name"
        changeset_type = "CREATE"

        mock_client = mock.Mock()
        mock_deployer = Deployer(mock_client)
        mock_waiter = mock.Mock()
        mock_client.get_waiter.return_value = mock_waiter

        waiter_error = botocore.exceptions.WaiterError(
            name="name", reason="reason", last_response={}
        )
        mock_waiter.wait.side_effect = waiter_error

        with self.assertRaises(exceptions.DeployFailedError):
            mock_deployer.wait_for_execute(stack_name, changeset_type)

        waiter_config = {
            'Delay': 30,
            'MaxAttempts': 120,
        }
        mock_waiter.wait.assert_called_once_with(
            StackName=stack_name, WaiterConfig=waiter_config
        )

        mock_client.get_waiter.assert_called_once_with("stack_create_complete")


def make_stack_obj(stack_name, status="CREATE_COMPLETE"):
    return {
        "StackId": stack_name,
        "StackName": stack_name,
        "CreationTime": "2013-08-23T01:02:15.422Z",
        "StackStatus": status,
    }



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import platform
import shutil
import socket
import stat
import tempfile

from botocore.exceptions import ClientError

from awscli.customizations.s3.filegenerator import (
    FileDecodingError,
    FileGenerator,
    FileStat,
    is_readable,
    is_special_file,
)
from awscli.customizations.s3.utils import EPOCH_TIME, get_file_stat
from awscli.testutils import (
    BaseAWSCommandParamsTest,
    FileCreator,
    mock,
    skip_if_windows,
    unittest,
)
from tests.unit.customizations.s3 import (
    clean_loc_files,
    compare_files,
    make_loc_files,
)


@skip_if_windows('Special files only supported on mac/linux')
class TestIsSpecialFile(unittest.TestCase):
    def setUp(self):
        self.files = FileCreator()
        self.filename = 'foo'

    def tearDown(self):
        self.files.remove_all()

    def test_is_character_device(self):
        file_path = os.path.join(self.files.rootdir, self.filename)
        self.files.create_file(self.filename, contents='')
        with mock.patch('stat.S_ISCHR') as mock_class:
            mock_class.return_value = True
            self.assertTrue(is_special_file(file_path))

    def test_is_block_device(self):
        file_path = os.path.join(self.files.rootdir, self.filename)
        self.files.create_file(self.filename, contents='')
        with mock.patch('stat.S_ISBLK') as mock_class:
            mock_class.return_value = True
            self.assertTrue(is_special_file(file_path))

    def test_is_fifo(self):
        file_path = os.path.join(self.files.rootdir, self.filename)
        mode = 0o600 | stat.S_IFIFO
        os.mknod(file_path, mode)
        self.assertTrue(is_special_file(file_path))

    def test_is_socket(self):
        file_path = os.path.join(self.files.rootdir, self.filename)
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.bind(file_path)
        self.assertTrue(is_special_file(file_path))


class TestIsReadable(unittest.TestCase):
    def setUp(self):
        self.files = FileCreator()
        self.filename = 'foo'
        self.full_path = os.path.join(self.files.rootdir, self.filename)

    def tearDown(self):
        self.files.remove_all()

    def test_unreadable_file(self):
        self.files.create_file(self.filename, contents="foo")
        open_function = 'awscli.customizations.s3.filegenerator._open'
        with mock.patch(open_function) as mock_class:
            mock_class.side_effect = OSError()
            self.assertFalse(is_readable(self.full_path))

    def test_unreadable_directory(self):
        os.mkdir(self.full_path)
        with mock.patch('os.listdir') as mock_class:
            mock_class.side_effect = OSError()
            self.assertFalse(is_readable(self.full_path))


class LocalFileGeneratorTest(unittest.TestCase):
    def setUp(self):
        self.client = None
        self.file_creator = FileCreator()
        self.files = make_loc_files(self.file_creator)
        self.local_file = self.files[0]
        self.local_dir = self.files[3] + os.sep

    def tearDown(self):
        clean_loc_files(self.file_creator)

    def test_local_file(self):
        """
        Generate a single local file.
        """
        input_local_file = {
            'src': {'path': self.local_file, 'type': 'local'},
            'dest': {'path': 'bucket/text1.txt', 'type': 's3'},
            'dir_op': False,
            'use_src_name': False,
        }
        params = {'region': 'us-east-1'}
        files = FileGenerator(self.client, '').call(input_local_file)
        result_list = []
        for filename in files:
            result_list.append(filename)
        size, last_update = get_file_stat(self.local_file)
        file_stat = FileStat(
            src=self.local_file,
            dest='bucket/text1.txt',
            compare_key='text1.txt',
            size=size,
            last_update=last_update,
            src_type='local',
            dest_type='s3',
            operation_name='',
        )
        ref_list = [file_stat]
        self.assertEqual(len(result_list), len(ref_list))
        for i in range(len(result_list)):
            compare_files(result_list[i], ref_list[i])

    def test_local_directory(self):
        """
        Generate an entire local directory.
        """
        input_local_dir = {
            'src': {'path': self.local_dir, 'type': 'local'},
            'dest': {'path': 'bucket/', 'type': 's3'},
            'dir_op': True,
            'use_src_name': True,
        }
        params = {'region': 'us-east-1'}
        files = FileGenerator(self.client, '').call(input_local_dir)
        result_list = []
        for filename in files:
            result_list.append(filename)
        size, last_update = get_file_stat(self.local_file)
        file_stat = FileStat(
            src=self.local_file,
            dest='bucket/text1.txt',
            compare_key='text1.txt',
            size=size,
            last_update=last_update,
            src_type='local',
            dest_type='s3',
            operation_name='',
        )
        path = self.local_dir + 'another_directory' + os.sep + 'text2.txt'
        size, last_update = get_file_stat(path)
        file_stat2 = FileStat(
            src=path,
            dest='bucket/another_directory/text2.txt',
            compare_key='another_directory/text2.txt',
            size=size,
            last_update=last_update,
            src_type='local',
            dest_type='s3',
            operation_name='',
        )
        ref_list = [file_stat2, file_stat]
        self.assertEqual(len(result_list), len(ref_list))
        for i in range(len(result_list)):
            compare_files(result_list[i], ref_list[i])


@skip_if_windows('Symlink tests only supported on mac/linux')
class TestIgnoreFilesLocally(unittest.TestCase):
    """
    This class tests the ability to ignore particular files.  This includes
    skipping symlink when desired.
    """

    def setUp(self):
        self.client = None
        self.files = FileCreator()

    def tearDown(self):
        self.files.remove_all()

    def test_warning(self):
        path = os.path.join(self.files.rootdir, 'badsymlink')
        os.symlink('non-existent-file', path)
        filegenerator = FileGenerator(self.client, '', True)
        self.assertTrue(filegenerator.should_ignore_file(path))

    def test_skip_symlink(self):
        filename = 'foo.txt'
        self.files.create_file(
            os.path.join(self.files.rootdir, filename),
            contents='foo.txt contents',
        )
        sym_path = os.path.join(self.files.rootdir, 'symlink')
        os.symlink(filename, sym_path)
        filegenerator = FileGenerator(self.client, '', False)
        self.assertTrue(filegenerator.should_ignore_file(sym_path))

    def test_no_skip_symlink(self):
        filename = 'foo.txt'
        path = self.files.create_file(
            os.path.join(self.files.rootdir, filename),
            contents='foo.txt contents',
        )
        sym_path = os.path.join(self.files.rootdir, 'symlink')
        os.symlink(path, sym_path)
        filegenerator = FileGenerator(self.client, '', True)
        self.assertFalse(filegenerator.should_ignore_file(sym_path))
        self.assertFalse(filegenerator.should_ignore_file(path))

    def test_no_skip_symlink_dir(self):
        filename = 'dir'
        path = os.path.join(self.files.rootdir, 'dir/')
        os.mkdir(path)
        sym_path = os.path.join(self.files.rootdir, 'symlink')
        os.symlink(path, sym_path)
        filegenerator = FileGenerator(self.client, '', True)
        self.assertFalse(filegenerator.should_ignore_file(sym_path))
        self.assertFalse(filegenerator.should_ignore_file(path))


class TestThrowsWarning(unittest.TestCase):
    def setUp(self):
        self.files = FileCreator()
        self.root = self.files.rootdir
        self.client = None

    def tearDown(self):
        self.files.remove_all()

    def test_no_warning(self):
        file_gen = FileGenerator(self.client, '', False)
        self.files.create_file("foo.txt", contents="foo")
        full_path = os.path.join(self.root, "foo.txt")
        return_val = file_gen.triggers_warning(full_path)
        self.assertFalse(return_val)
        self.assertTrue(file_gen.result_queue.empty())

    def test_no_exists(self):
        file_gen = FileGenerator(self.client, '', False)
        filename = os.path.join(self.root, 'file')
        return_val = file_gen.triggers_warning(filename)
        self.assertTrue(return_val)
        warning_message = file_gen.result_queue.get()
        self.assertEqual(
            warning_message.message,
            ("warning: Skipping file %s. File does not exist." % filename),
        )

    def test_no_read_access(self):
        file_gen = FileGenerator(self.client, '', False)
        self.files.create_file("foo.txt", contents="foo")
        full_path = os.path.join(self.root, "foo.txt")
        open_function = 'awscli.customizations.s3.filegenerator._open'
        with mock.patch(open_function) as mock_class:
            mock_class.side_effect = OSError()
            return_val = file_gen.triggers_warning(full_path)
            self.assertTrue(return_val)
        warning_message = file_gen.result_queue.get()
        self.assertEqual(
            warning_message.message,
            (
                "warning: Skipping file %s. File/Directory is "
                "not readable." % full_path
            ),
        )

    @skip_if_windows('Special files only supported on mac/linux')
    def test_is_special_file_warning(self):
        file_gen = FileGenerator(self.client, '', False)
        file_path = os.path.join(self.files.rootdir, 'foo')
        # Use socket for special file.
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.bind(file_path)
        return_val = file_gen.triggers_warning(file_path)
        self.assertTrue(return_val)
        warning_message = file_gen.result_queue.get()
        self.assertEqual(
            warning_message.message,
            (
                "warning: Skipping file %s. File is character "
                "special device, block special device, FIFO, or "
                "socket." % file_path
            ),
        )


@skip_if_windows('Symlink tests only supported on mac/linux')
class TestSymlinksIgnoreFiles(unittest.TestCase):
    """
    This class tests the ability to list out the correct local files
    depending on if symlinks are being followed.  Also tests to ensure
    broken symlinks fail.
    """

    def setUp(self):
        self.client = None
        self.files = FileCreator()
        # List of local filenames.
        self.filenames = []
        self.root = self.files.rootdir
        self.bucket = 'bucket/'
        filename_1 = self.files.create_file(
            'foo.txt', contents='foo.txt contents'
        )
        self.filenames.append(filename_1)
        nested_dir = os.path.join(self.root, 'realfiles')
        os.mkdir(nested_dir)
        filename_2 = self.files.create_file(
            os.path.join(nested_dir, 'bar.txt'), contents='bar.txt contents'
        )
        self.filenames.append(filename_2)
        # Names of symlinks.
        self.symlinks = []
        # Names of files if symlinks are followed.
        self.symlink_files = []
        # Create symlink to file foo.txt.
        symlink_1 = os.path.join(self.root, 'symlink_1')
        os.symlink(filename_1, symlink_1)
        self.symlinks.append(symlink_1)
        self.symlink_files.append(symlink_1)
        # Create a symlink to a file that does not exist.
        symlink_2 = os.path.join(self.root, 'symlink_2')
        os.symlink('non-existent-file', symlink_2)
        self.symlinks.append(symlink_2)
        # Create a symlink to directory realfiles
        symlink_3 = os.path.join(self.root, 'symlink_3')
        os.symlink(nested_dir, symlink_3)
        self.symlinks.append(symlink_3)
        self.symlink_files.append(os.path.join(symlink_3, 'bar.txt'))

    def tearDown(self):
        self.files.remove_all()

    def test_no_follow_symlink(self):
        abs_root = str(os.path.abspath(self.root) + os.sep)
        input_local_dir = {
            'src': {'path': abs_root, 'type': 'local'},
            'dest': {'path': self.bucket, 'type': 's3'},
            'dir_op': True,
            'use_src_name': True,
        }
        file_stats = FileGenerator(self.client, '', False).call(
            input_local_dir
        )
        self.filenames.sort()
        result_list = []
        for file_stat in file_stats:
            result_list.append(getattr(file_stat, 'src'))
        self.assertEqual(len(result_list), len(self.filenames))
        # Just check to make sure the right local files are generated.
        for i in range(len(result_list)):
            filename = str(os.path.abspath(self.filenames[i]))
            self.assertEqual(result_list[i], filename)

    def test_warn_bad_symlink(self):
        """
        This tests to make sure it fails when following bad symlinks.
        """
        abs_root = str(os.path.abspath(self.root) + os.sep)
        input_local_dir = {
            'src': {'path': abs_root, 'type': 'local'},
            'dest': {'path': self.bucket, 'type': 's3'},
            'dir_op': True,
            'use_src_name': True,
        }
        file_stats = FileGenerator(self.client, '', True).call(input_local_dir)
        file_gen = FileGenerator(self.client, '', True)
        file_stats = file_gen.call(input_local_dir)
        all_filenames = self.filenames + self.symlink_files
        all_filenames.sort()
        result_list = []
        for file_stat in file_stats:
            result_list.append(getattr(file_stat, 'src'))
        self.assertEqual(len(result_list), len(all_filenames))
        # Just check to make sure the right local files are generated.
        for i in range(len(result_list)):
            filename = str(os.path.abspath(all_filenames[i]))
            self.assertEqual(result_list[i], filename)
        self.assertFalse(file_gen.result_queue.empty())

    def test_follow_symlink(self):
        # First remove the bad symlink.
        os.remove(os.path.join(self.root, 'symlink_2'))
        abs_root = str(os.path.abspath(self.root) + os.sep)
        input_local_dir = {
            'src': {'path': abs_root, 'type': 'local'},
            'dest': {'path': self.bucket, 'type': 's3'},
            'dir_op': True,
            'use_src_name': True,
        }
        file_stats = FileGenerator(self.client, '', True).call(input_local_dir)
        all_filenames = self.filenames + self.symlink_files
        all_filenames.sort()
        result_list = []
        for file_stat in file_stats:
            result_list.append(getattr(file_stat, 'src'))
        self.assertEqual(len(result_list), len(all_filenames))
        # Just check to make sure the right local files are generated.
        for i in range(len(result_list)):
            filename = str(os.path.abspath(all_filenames[i]))
            self.assertEqual(result_list[i], filename)


class TestListFilesLocally(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self.directory = str(tempfile.mkdtemp())

    def tearDown(self):
        shutil.rmtree(self.directory)

    @mock.patch('os.listdir')
    def test_error_raised_on_decoding_error(self, listdir_mock):
        # On Python3, sys.getdefaultencoding
        file_generator = FileGenerator(None, None, None)
        # utf-8 encoding for U+2713.
        listdir_mock.return_value = [b'\xe2\x9c\x93']
        list(file_generator.list_files(self.directory, dir_op=True))
        # Ensure the message was added to the result queue and is
        # being skipped.
        self.assertFalse(file_generator.result_queue.empty())
        warning_message = file_generator.result_queue.get()
        self.assertIn("warning: Skipping file ", warning_message.message)
        self.assertIn(
            "Please check your locale settings.", warning_message.message
        )

    def test_list_files_is_in_sorted_order(self):
        p = os.path.join
        open(p(self.directory, 'test-123.txt'), 'w').close()
        open(p(self.directory, 'test-321.txt'), 'w').close()
        open(p(self.directory, 'test123.txt'), 'w').close()
        open(p(self.directory, 'test321.txt'), 'w').close()
        os.mkdir(p(self.directory, 'test'))
        open(p(self.directory, 'test', 'foo.txt'), 'w').close()

        file_generator = FileGenerator(None, None, None)
        values = list(
            el[0]
            for el in file_generator.list_files(self.directory, dir_op=True)
        )
        ref_vals = list(
            sorted(values, key=lambda items: items.replace(os.sep, '/'))
        )
        self.assertEqual(values, ref_vals)

    @mock.patch('awscli.customizations.s3.filegenerator.get_file_stat')
    def test_list_files_with_invalid_timestamp(self, stat_mock):
        stat_mock.return_value = 9, None
        open(os.path.join(self.directory, 'test'), 'w').close()
        file_generator = FileGenerator(None, None, None)
        value = list(file_generator.list_files(self.directory, dir_op=True))[0]
        self.assertIs(value[1]['LastModified'], EPOCH_TIME)

    def test_list_local_files_with_unicode_chars(self):
        p = os.path.join
        open(p(self.directory, 'a'), 'w').close()
        open(p(self.directory, 'a\u0300'), 'w').close()
        open(p(self.directory, 'a\u0300-1'), 'w').close()
        open(p(self.directory, 'a\u03001'), 'w').close()
        open(p(self.directory, 'z'), 'w').close()
        open(p(self.directory, '\u00e6'), 'w').close()
        os.mkdir(p(self.directory, 'a\u0300a'))
        open(p(self.directory, 'a\u0300a', 'a'), 'w').close()
        open(p(self.directory, 'a\u0300a', 'z'), 'w').close()
        open(p(self.directory, 'a\u0300a', '\u00e6'), 'w').close()

        file_generator = FileGenerator(None, None, None)
        values = list(
            el[0]
            for el in file_generator.list_files(self.directory, dir_op=True)
        )
        expected_order = [
            os.path.join(self.directory, el)
            for el in [
                "a",
                "a\u0300",
                "a\u0300-1",
                "a\u03001",
                "a\u0300a%sa" % os.path.sep,
                "a\u0300a%sz" % os.path.sep,
                "a\u0300a%s\u00e6" % os.path.sep,
                "z",
                "\u00e6",
            ]
        ]
        self.assertEqual(values, expected_order)


class TestNormalizeSort(unittest.TestCase):
    def test_normalize_sort(self):
        names = [
            'xyz123456789',
            'xyz1' + os.path.sep + 'test',
            'xyz' + os.path.sep + 'test',
        ]
        ref_names = [names[2], names[1], names[0]]
        filegenerator = FileGenerator(None, None, None)
        filegenerator.normalize_sort(names, os.path.sep, '/')
        for i in range(len(ref_names)):
            self.assertEqual(ref_names[i], names[i])

    def test_normalize_sort_backslash(self):
        names = ['xyz123456789', 'xyz1\\test', 'xyz\\test']
        ref_names = [names[2], names[1], names[0]]
        filegenerator = FileGenerator(None, None, None)
        filegenerator.normalize_sort(names, '\\', '/')
        for i in range(len(ref_names)):
            self.assertEqual(ref_names[i], names[i])


class S3FileGeneratorTest(BaseAWSCommandParamsTest):
    def setUp(self):
        super(S3FileGeneratorTest, self).setUp()
        self.client = self.driver.session.create_client('s3')
        self.bucket = 'foo'
        self.file1 = self.bucket + '/' + 'text1.txt'
        self.file2 = self.bucket + '/' + 'another_directory/text2.txt'

    def test_s3_file(self):
        """
        Generate a single s3 file
        Note: Size and last update are not tested because s3 generates them.
        """
        input_s3_file = {
            'src': {'path': self.file1, 'type': 's3'},
            'dest': {'path': 'text1.txt', 'type': 'local'},
            'dir_op': False,
            'use_src_name': False,
        }
        params = {'region': 'us-east-1'}
        self.parsed_responses = [
            {
                "ETag": "abcd",
                "ContentLength": 100,
                "LastModified": "2014-01-09T20:45:49.000Z",
            }
        ]
        self.patch_make_request()

        file_gen = FileGenerator(self.client, '')
        files = file_gen.call(input_s3_file)
        result_list = []
        for filename in files:
            result_list.append(filename)
        file_stat = FileStat(
            src=self.file1,
            dest='text1.txt',
            compare_key='text1.txt',
            size=result_list[0].size,
            last_update=result_list[0].last_update,
            src_type='s3',
            dest_type='local',
            operation_name='',
        )

        ref_list = [file_stat]
        self.assertEqual(len(result_list), len(ref_list))
        for i in range(len(result_list)):
            compare_files(result_list[i], ref_list[i])

    def test_s3_single_file_404(self):
        """
        Test the error message for a 404 ClientError for a single file listing
        """
        input_s3_file = {
            'src': {'path': self.file1, 'type': 's3'},
            'dest': {'path': 'text1.txt', 'type': 'local'},
            'dir_op': False,
            'use_src_name': False,
        }
        params = {'region': 'us-east-1'}
        self.client = mock.Mock()
        self.client.head_object.side_effect = ClientError(
            {'Error': {'Code': '404', 'Message': 'Not Found'}},
            'HeadObject',
        )
        file_gen = FileGenerator(self.client, '')
        files = file_gen.call(input_s3_file)
        # The error should include 404 and should include the key name.
        with self.assertRaisesRegex(ClientError, '404.*text1.txt'):
            list(files)

    def test_s3_single_file_delete(self):
        input_s3_file = {
            'src': {'path': self.file1, 'type': 's3'},
            'dest': {'path': '', 'type': 'local'},
            'dir_op': False,
            'use_src_name': True,
        }
        self.client = mock.Mock()
        file_gen = FileGenerator(self.client, 'delete')
        result_list = list(file_gen.call(input_s3_file))
        self.assertEqual(len(result_list), 1)
        compare_files(
            result_list[0],
            FileStat(
                src=self.file1,
                dest='text1.txt',
                compare_key='text1.txt',
                size=None,
                last_update=None,
                src_type='s3',
                dest_type='local',
                operation_name='delete',
            ),
        )
        self.client.head_object.assert_not_called()

    def test_s3_single_file_sets_checksum_mode_when_supported(self):
        input_s3_file = {
            'src': {'path': self.file1, 'type': 's3'},
            'dest': {'path': 'text1.txt', 'type': 'local'},
            'dir_op': False,
            'use_src_name': False,
        }
        self.client = mock.Mock()
        self.client.meta.config.response_checksum_validation = 'when_supported'
        self.client.head_object.return_value = {
            'ContentLength': 100,
            'LastModified': '2014-01-09T20:45:49.000Z',
            'ETag': '"abcd"',
        }
        file_gen = FileGenerator(self.client, '')
        list(file_gen.call(input_s3_file))
        call_kwargs = self.client.head_object.call_args[1]
        self.assertEqual(call_kwargs['ChecksumMode'], 'ENABLED')

    def test_s3_single_file_no_checksum_mode_when_required(self):
        input_s3_file = {
            'src': {'path': self.file1, 'type': 's3'},
            'dest': {'path': 'text1.txt', 'type': 'local'},
            'dir_op': False,
            'use_src_name': False,
        }
        self.client = mock.Mock()
        self.client.meta.config.response_checksum_validation = 'when_required'
        self.client.head_object.return_value = {
            'ContentLength': 100,
            'LastModified': '2014-01-09T20:45:49.000Z',
            'ETag': '"abcd"',
        }
        file_gen = FileGenerator(self.client, '')
        list(file_gen.call(input_s3_file))
        call_kwargs = self.client.head_object.call_args[1]
        self.assertNotIn('ChecksumMode', call_kwargs)

    def test_s3_single_file_explicit_checksum_mode_overrides(self):
        input_s3_file = {
            'src': {'path': self.file1, 'type': 's3'},
            'dest': {'path': 'text1.txt', 'type': 'local'},
            'dir_op': False,
            'use_src_name': False,
        }
        self.client = mock.Mock()
        self.client.meta.config.response_checksum_validation = 'when_required'
        self.client.head_object.return_value = {
            'ContentLength': 100,
            'LastModified': '2014-01-09T20:45:49.000Z',
            'ETag': '"abcd"',
        }
        file_gen = FileGenerator(
            self.client,
            '',
            request_parameters={
                'HeadObject': {'ChecksumMode': 'ENABLED'},
            },
        )
        list(file_gen.call(input_s3_file))
        call_kwargs = self.client.head_object.call_args[1]
        self.assertEqual(call_kwargs['ChecksumMode'], 'ENABLED')

    def test_s3_directory(self):
        """
        Generates s3 files under a common prefix. Also it ensures that
        zero size files are ignored.
        Note: Size and last update are not tested because s3 generates them.
        """
        input_s3_file = {
            'src': {'path': self.bucket + '/', 'type': 's3'},
            'dest': {'path': '', 'type': 'local'},
            'dir_op': True,
            'use_src_name': True,
        }
        params = {'region': 'us-east-1'}
        files = FileGenerator(self.client, '').call(input_s3_file)

        self.parsed_responses = [
            {
                "CommonPrefixes": [],
                "Contents": [
                    {
                        "Key": "another_directory/text2.txt",
                        "Size": 100,
                        "LastModified": "2014-01-09T20:45:49.000Z",
                    },
                    {
                        "Key": "text1.txt",
                        "Size": 10,
                        "LastModified": "2013-01-09T20:45:49.000Z",
                    },
                ],
            }
        ]
        self.patch_make_request()
        result_list = []
        for filename in files:
            result_list.append(filename)
        file_stat = FileStat(
            src=self.file2,
            dest='another_directory' + os.sep + 'text2.txt',
            compare_key='another_directory/text2.txt',
            size=result_list[0].size,
            last_update=result_list[0].last_update,
            src_type='s3',
            dest_type='local',
            operation_name='',
        )
        file_stat2 = FileStat(
            src=self.file1,
            dest='text1.txt',
            compare_key='text1.txt',
            size=result_list[1].size,
            last_update=result_list[1].last_update,
            src_type='s3',
            dest_type='local',
            operation_name='',
        )

        ref_list = [file_stat, file_stat2]
        self.assertEqual(len(result_list), len(ref_list))
        for i in range(len(result_list)):
            compare_files(result_list[i], ref_list[i])

    def test_s3_delete_directory(self):
        """
        Generates s3 files under a common prefix. Also it ensures that
        the directory itself is included because it is a delete command
        Note: Size and last update are not tested because s3 generates them.
        """
        input_s3_file = {
            'src': {'path': self.bucket + '/', 'type': 's3'},
            'dest': {'path': '', 'type': 'local'},
            'dir_op': True,
            'use_src_name': True,
        }
        self.parsed_responses = [
            {
                "CommonPrefixes": [],
                "Contents": [
                    {
                        "Key": "another_directory/",
                        "Size": 0,
                        "LastModified": "2012-01-09T20:45:49.000Z",
                    },
                    {
                        "Key": "another_directory/text2.txt",
                        "Size": 100,
                        "LastModified": "2014-01-09T20:45:49.000Z",
                    },
                    {
                        "Key": "text1.txt",
                        "Size": 10,
                        "LastModified": "2013-01-09T20:45:49.000Z",
                    },
                ],
            }
        ]
        self.patch_make_request()
        files = FileGenerator(self.client, 'delete').call(input_s3_file)
        result_list = []
        for filename in files:
            result_list.append(filename)

        file_stat1 = FileStat(
            src=self.bucket + '/another_directory/',
            dest='another_directory' + os.sep,
            compare_key='another_directory/',
            size=result_list[0].size,
            last_update=result_list[0].last_update,
            src_type='s3',
            dest_type='local',
            operation_name='delete',
        )
        file_stat2 = FileStat(
            src=self.file2,
            dest='another_directory' + os.sep + 'text2.txt',
            compare_key='another_directory/text2.txt',
            size=result_list[1].size,
            last_update=result_list[1].last_update,
            src_type='s3',
            dest_type='local',
            operation_name='delete',
        )
        file_stat3 = FileStat(
            src=self.file1,
            dest='text1.txt',
            compare_key='text1.txt',
            size=result_list[2].size,
            last_update=result_list[2].last_update,
            src_type='s3',
            dest_type='local',
            operation_name='delete',
        )

        ref_list = [file_stat1, file_stat2, file_stat3]
        self.assertEqual(len(result_list), len(ref_list))
        for i in range(len(result_list)):
            compare_files(result_list[i], ref_list[i])


if __name__ == "__main__":
    unittest.main()



import json
import math
import os
import shutil
import sys
import time

import psutil

from awscli.clidriver import AWSCLIEntryPoint, create_clidriver
from scripts.performance import BaseBenchmarkSuite
from scripts.performance.simple_stubbed_tests import (
    JSONStubbedBenchmarkSuite,
)


class Metric:
    def __init__(self, description, unit, value):
        self.description = description
        self.unit = unit
        self.value = value


class BenchmarkResultsSerializer:
    """
    A class that serializes the execution results of a performance test case.
    """

    def __init__(self):
        self._summarizer = Summarizer()
        self._benchmark_results = {}

    def add_execution_results(self, case, samples, execution_results):
        """
        Store a performance test case's execution result.
        """
        summarized_results = self._summarizer.summarize(
            samples, execution_results
        )
        for metric, val in summarized_results.items():
            key = f'{case["name"]}.{metric}'
            if key not in self._benchmark_results:
                self._benchmark_results[key] = {
                    'name': key,
                    'description': val.description,
                    'unit': val.unit,
                    'dimensions': case.get('dimensions', []),
                    'measurements': [],
                }
            self._benchmark_results[key]['measurements'].append(val.value)

    def get_processed_results(self):
        """
        Returns a list of dictionaries representing all stored execution
        results. The key-value pairs will be converted to JSON and displayed as part
        of the final output.
        """
        return list(self._benchmark_results.values())

    def reset(self):
        """
        Resets the stored list of execution results.
        """
        self._benchmark_results.clear()


class Summarizer:
    DATA_INDEX_IN_ROW = {'time': 0, 'memory': 1, 'cpu': 2}

    def __init__(self):
        self._start_time = None
        self._end_time = None
        self._samples = []
        self._sums = {
            'memory': 0.0,
            'cpu': 0.0,
        }

    def summarize(self, samples, worker_results):
        """
        Processes benchmark data from samples and the output of the benchmark
        worker.
        """
        self._samples = samples
        self._validate_samples(samples)
        for idx, sample in enumerate(samples):
            # If the sample is the first one, collect the start time.
            if idx == 0:
                self._start_time = self._get_time(sample)
            self.process_data_sample(sample)
        self._end_time = self._get_time(samples[-1])
        metrics = self._finalize_processed_data_for_file(
            samples, worker_results
        )
        return metrics

    def _validate_samples(self, samples):
        if not samples:
            raise RuntimeError(
                'Benchmark samples could not be processed. '
                'The samples list is empty'
            )

    def process_data_sample(self, sample):
        self._add_to_sums('memory', sample['memory'])
        self._add_to_sums('cpu', sample['cpu'])

    def _finalize_processed_data_for_file(self, samples, worker_results):
        # compute percentiles
        self._samples.sort(key=self._get_memory)
        memory_p50 = self._compute_metric_percentile(50, 'memory')
        memory_p95 = self._compute_metric_percentile(95, 'memory')
        self._samples.sort(key=self._get_cpu)
        cpu_p50 = self._compute_metric_percentile(50, 'cpu')
        cpu_p95 = self._compute_metric_percentile(95, 'cpu')
        max_memory = max(samples, key=self._get_memory)['memory']
        max_cpu = max(samples, key=self._get_cpu)['cpu']
        # format computed statistics
        metrics = {
            'mean.run.memory': Metric(
                'Mean memory usage of a single command execution.',
                'Bytes',
                self._sums['memory'] / len(samples),
            ),
            'mean.run.cpu': Metric(
                'Mean CPU usage of a single command execution.',
                'Percent',
                self._sums['cpu'] / len(samples),
            ),
            'peak.run.memory': Metric(
                'Peak memory usage of a single command execution.',
                'Bytes',
                max_memory,
            ),
            'peak.run.cpu': Metric(
                'Peak CPU usage of a single command execution.',
                'Percent',
                max_cpu,
            ),
            'p50.run.memory': Metric(
                'p50 memory usage of a single command execution.',
                'Bytes',
                memory_p50,
            ),
            'p95.run.memory': Metric(
                'p95 memory usage of a single command execution.',
                'Bytes',
                memory_p95,
            ),
            'p50.run.cpu': Metric(
                'p50 CPU usage of a single command execution.',
                'Percent',
                cpu_p50,
            ),
            'p95.run.cpu': Metric(
                'p95 CPU usage of a single command execution.',
                'Percent',
                cpu_p95,
            ),
            'run.time': Metric(
                'Total running time of the Python process executing the CLI command.',
                'Seconds',
                worker_results['end_time'] - worker_results['start_time'],
            ),
            'pre.marshal.time': Metric(
                'Elapsed time from the start of the Python process until just '
                'before the HTTP request is created.',
                'Seconds',
                worker_results['first_client_invocation_time']
                - worker_results['start_time'],
            ),
        }
        # reset data state
        self._samples.clear()
        self._sums = self._sums.fromkeys(self._sums, 0.0)
        return metrics

    def _compute_metric_percentile(self, percentile, name):
        num_samples = len(self._samples)
        p_idx = math.ceil(percentile * num_samples / 100) - 1
        return self._samples[p_idx][name]

    def _get_time(self, sample):
        return sample['time']

    def _get_memory(self, sample):
        return sample['memory']

    def _get_cpu(self, sample):
        return sample['cpu'] / 100

    def _add_to_sums(self, name, data_point):
        self._sums[name] += data_point


class ProcessBenchmarker:
    """
    Periodically samples CPU and memory usage of a process given its pid.
    These measurements are sampled until the process is no longer running.
    """

    def benchmark_process(self, pid, data_interval):
        parent_pid = os.getpid()
        try:
            # Benchmark the process where the script is being run.
            return self._run_benchmark(pid, data_interval)
        except KeyboardInterrupt:
            # If there is an interrupt, then try to clean everything up.
            proc = psutil.Process(parent_pid)
            procs = proc.children(recursive=True)

            for child in procs:
                child.terminate()

            gone, alive = psutil.wait_procs(procs, timeout=1)
            for child in alive:
                child.kill()
            raise

    def _run_benchmark(self, pid, data_interval):
        process_to_measure = psutil.Process(pid)
        samples = []

        while process_to_measure.is_running():
            if process_to_measure.status() == psutil.STATUS_ZOMBIE:
                break
            time.sleep(data_interval)
            try:
                # Collect the memory and cpu usage.
                memory_used = process_to_measure.memory_info().rss
                cpu_percent = process_to_measure.cpu_percent()
            except (
                psutil.AccessDenied,
                psutil.ZombieProcess,
                psutil.NoSuchProcess,
            ):
                # Trying to get process information from a closed or
                # zombie process will result in corresponding exceptions.
                break
            # Determine the lapsed time for bookkeeping
            current_time = time.time()
            samples.append(
                {
                    "time": current_time,
                    "memory": memory_used,
                    "cpu": cpu_percent,
                }
            )
        return samples


class BenchmarkHarness:
    BENCHMARK_SUITES = [JSONStubbedBenchmarkSuite]

    """
    Orchestrates running benchmarks in isolated, configurable environments.
    """

    def __init__(self, results_processor=BenchmarkResultsSerializer()):
        self._results_processor = results_processor

    def _run_command_with_metric_hooks(self, cmd, out_file):
        """
        Runs a CLI command and logs CLI-specific metrics to a file.
        """
        first_client_invocation_time = None
        start_time = time.time()
        driver = create_clidriver()
        event_emitter = driver.session.get_component('event_emitter')

        def _log_invocation_time(params, request_signer, model, **kwargs):
            nonlocal first_client_invocation_time
            if first_client_invocation_time is None:
                first_client_invocation_time = time.time()

        event_emitter.register_last(
            'before-call',
            _log_invocation_time,
            'benchmarks.log-invocation-time',
        )

        rc = AWSCLIEntryPoint(driver).main(cmd)
        end_time = time.time()

        # write the collected metrics to a file
        with open(out_file, 'w') as metrics_f:
            metrics_f.write(
                json.dumps(
                    {
                        'return_code': rc,
                        'start_time': start_time,
                        'end_time': end_time,
                        'first_client_invocation_time': first_client_invocation_time,
                    }
                )
            )

    def _run_isolated_benchmark(
        self,
        result_dir,
        iteration,
        benchmark,
        suite,
        process_benchmarker,
        args,
    ):
        """
        Runs a single iteration of one benchmark execution. Includes setting up
        the environment, running the benchmarked execution, formatting
        the results, and cleaning up the environment.
        """
        assets_path = os.path.join(result_dir, 'assets')
        metrics_path = os.path.join(assets_path, 'metrics.json')
        child_output_path = os.path.join(assets_path, 'output.txt')
        child_err_path = os.path.join(assets_path, 'err.txt')

        # setup for iteration of benchmark
        suite.begin_iteration(benchmark, result_dir, assets_path, iteration)
        os.chdir(result_dir)

        # fork a child process to run the command on.
        pid = os.fork()

        try:
            if pid == 0:
                with (
                    open(child_output_path, 'w') as out,
                    open(child_err_path, 'w') as err,
                ):
                    if not args.debug_dir:
                        # redirect standard output of the child process to a file
                        os.dup2(out.fileno(), sys.stdout.fileno())
                        os.dup2(err.fileno(), sys.stderr.fileno())
                    else:
                        with open(
                            os.path.abspath(
                                os.path.join(
                                    args.debug_dir,
                                    f'{benchmark["name"]}-{iteration}.txt',
                                )
                            ),
                            'w',
                        ) as f:
                            with open(
                                os.path.abspath(
                                    os.path.join(
                                        args.debug_dir,
                                        f'{benchmark["name"]}-{iteration}-err.txt',
                                    )
                                ),
                                'w',
                            ) as f_err:
                                os.dup2(f.fileno(), sys.stdout.fileno())
                                os.dup2(f_err.fileno(), sys.stderr.fileno())
                    # execute command on child process
                    self._run_command_with_metric_hooks(
                        benchmark['command'], metrics_path
                    )
                    # terminate the child process
                    os._exit(0)

            # benchmark child process from parent process until child becomes zombie
            samples = process_benchmarker.benchmark_process(
                pid, args.data_interval
            )

            # reap the child process and error on unsuccessful return codes
            _, status = os.waitpid(pid, 0)
            if status != 0:
                raise RuntimeError(
                    f'Child process execution failed: status code {status}'
                )

            # load child-collected metrics
            if not os.path.exists(metrics_path):
                raise RuntimeError(
                    'Child process execution failed: output file not found.'
                )
            worker_results = json.load(open(metrics_path))

            # raise error if CLI execution unsuccessful.
            # this is different from the process return code checked above,
            # because the process can succeed while the CLI execution failed
            if (rc := worker_results['return_code']) != 0:
                with open(child_err_path) as err:
                    raise RuntimeError(
                        f'CLI execution failed: return code {rc}.\n'
                        f'Error: {err.read()}'
                    )

            # summarize benchmark results and process summary
            return samples, worker_results
        finally:
            suite.end_iteration(benchmark, iteration)
            shutil.rmtree(result_dir, ignore_errors=True)
            os.makedirs(result_dir, 0o777)

    def get_test_suites(self, args):
        """
        Returns all test suites that should be executed by the default
        performance test runner.
        """
        return [suite() for suite in BenchmarkHarness.BENCHMARK_SUITES]

    def run_benchmarks(self, cases, args):
        """
        Orchestrates benchmarking via the supplied list of performance test
        cases.
        """
        summaries = {'results': []}
        result_dir = args.result_dir
        process_benchmarker = ProcessBenchmarker()

        if os.path.exists(result_dir):
            shutil.rmtree(result_dir)
        os.makedirs(result_dir, 0o777)
        try:
            for suite, case in cases:
                for idx in range(args.num_iterations):
                    for cmd in case:
                        samples, execution_results = (
                            self._run_isolated_benchmark(
                                result_dir,
                                idx,
                                cmd,
                                suite,
                                process_benchmarker,
                                args,
                            )
                        )
                        self._results_processor.add_execution_results(
                            cmd, samples, execution_results
                        )
                summaries['results'].extend(
                    self._results_processor.get_processed_results()
                )
                self._results_processor.reset()
        finally:
            # final cleanup
            shutil.rmtree(result_dir, ignore_errors=True)
        print(json.dumps(summaries, indent=2))

    def run_benchmark_suite(self, suite: BaseBenchmarkSuite, args):
        """
        Orchestrates benchmarking a particular benchmark suite.
        """
        sequence_generators = suite.get_test_cases(args)
        self.run_benchmarks(sequence_generators, args)



import base64
from functools import partial

import pytest
from awscrt.crypto import EC
from botocore.utils import build_dpop_header

EC_PRIVATE_KEY_SEC1_BASE64 = (
    'MHcCAQEEIHjt7c+VnkIkN6RW7QgZPFNLb/9AZEhqSYYMtwrlLb3WoAoGCCqGSM49'
    'AwEHoUQDQgAEv2FjRpMtADMZ4zoZxshV9chEkembgzZnXSUNe+DA8dKqXN/7qTcZ'
    'jYJHKIi+Rn88zUGqCJo3DWF/X+ufVfdU2g=='
)


@pytest.fixture
def private_key():
    """Loads a fixed private key"""
    return EC.new_key_from_der_data(
        base64.b64decode(EC_PRIVATE_KEY_SEC1_BASE64)
    )


@pytest.fixture
def fake_build_dpop_header(private_key):
    """Overrides DPoP signing with frozen key, URI, and UID"""
    return partial(
        build_dpop_header,
        private_key=private_key,
        uri='https://foobar/v1/token',
        uid='uniqueid',
        ts=1760014198,
    )



# Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import os
import platform
import sys
from typing import List

import pytest
from build_system.utils import (
    ParseError,
    Requirement,
    UnmetDependenciesException,
    Utils,
    parse_requirements,
)

from tests.markers import if_windows, skip_if_windows


@pytest.fixture
def utils():
    return Utils()


@pytest.mark.parametrize(
    "lines,expected",
    [
        (
            "flit_core>=3.7.1,<3.7.2",
            Requirement("flit_core", ">=3.7.1", "<3.7.2"),
        ),
        (
            "colorama>=0.2.5,<0.4.4",
            Requirement("colorama", ">=0.2.5", "<0.4.4"),
        ),
        ("docutils>=0.10,<0.16", Requirement("docutils", ">=0.10", "<0.16")),
        (
            "ruamel_yaml>=0.15.0,<=0.17.21",
            Requirement("ruamel_yaml", ">=0.15.0", "<=0.17.21"),
        ),
        ("wcwidth<0.2.0", Requirement("wcwidth", "<0.2.0")),
        (
            "prompt-toolkit>=3.0.24,<3.0.29",
            Requirement("prompt-toolkit", ">=3.0.24", "<3.0.29"),
        ),
        ("distro>=1.5.0,<1.6.0", Requirement("distro", ">=1.5.0", "<1.6.0")),
        (
            "awscrt>=0.12.4,<=0.14.0",
            Requirement("awscrt", ">=0.12.4", "<=0.14.0"),
        ),
        (
            "python-dateutil>=2.1,<=2.8.2",
            Requirement("python-dateutil", ">=2.1", "<=2.8.2"),
        ),
        (
            "jmespath>=0.7.1,<1.1.0",
            Requirement("jmespath", ">=0.7.1", "<1.1.0"),
        ),
        ("urllib3>=1.25.4,<1.27", Requirement("urllib3", ">=1.25.4", "<1.27")),
        (
            ["urllib3>=1.\\\\", "25.4,<1.27"],
            Requirement("urllib3", ">=1.25.4", "<1.27"),
        ),
        ("#urllib3>=1.25.4,<1.27", None),
        (
            "urllib3>=1.25.4,<1.27 #foobarbaz",
            Requirement("urllib3", ">=1.25.4", "<1.27"),
        ),
        ("urllib3>=1.25.4,<1.27 ; python_version == 3.6", ParseError),
    ],
)
def test_parse_requirements(lines, expected):
    if isinstance(lines, str):
        lines = [lines]
    if isinstance(expected, Requirement):
        req = list(parse_requirements(lines))[0]
        assert req == expected
    elif expected is None:
        assert len(list(parse_requirements(lines))) == 0
    else:
        with pytest.raises(expected):
            list(parse_requirements(lines))


@pytest.mark.parametrize(
    "req,version,expected",
    [
        (Requirement("foo", "==1.0"), "1.0", True),
        (Requirement("foo", "==1.0"), "1.0.0", True),
        (Requirement("foo", "==1.0.0"), "1.0", True),
        (Requirement("foo", "==1.0"), "2.0", False),
        (Requirement("foo", ">=1.0", "<2.0"), "1.1", True),
        (Requirement("foo", ">=3.7.1", "<3.7.2"), "3.7.1", True),
        (Requirement("foo", ">=3.7.1", "<3.7.2"), "3.7.2", False),
        (Requirement("foo", ">=3.7.1", "<3.7.2"), "1.1.1", False),
        (Requirement("foo", ">=3.7.1", "<3.7.2"), "1.1.1", False),
        (Requirement("foo", ">=0.15.0", "<=0.17.21"), "0.14.0", False),
        (Requirement("foo", ">=0.15.0", "<=0.17.21"), "0.14.99999", False),
        (Requirement("foo", ">=0.15.0", "<=0.17.21"), "0.14.99999.123", False),
        (Requirement("foo", ">=0.15.0", "<=0.17.21"), "0.15.0", True),
        (Requirement("foo", ">=0.15.0", "<=0.17.21"), "0.16.0", True),
        (Requirement("foo", ">=0.15.0", "<=0.17.21"), "0.17.0", True),
        (Requirement("foo", ">=0.15.0", "<=0.17.21"), "0.17.21", True),
        (Requirement("foo", ">=0.15.0", "<=0.17.21"), "0.17.22", False),
        (Requirement("foo", ">=0.15.0", "<=0.17.21"), "0.18.0", False),
    ],
)
def test_requirement_ranges(req, version, expected):
    assert req.is_in_range(version) == expected


class TestUtils:
    def test_isdir(self, utils: Utils, tmp_path):
        file_path = tmp_path / "filename.txt"
        file_path.write_text("foo")

        assert utils.isdir(tmp_path) is True
        assert utils.isdir(file_path) is False

    def test_path_exists(self, utils: Utils, tmp_path):
        file_path = tmp_path / "filename.txt"
        file_path.write_text("foo")
        fake_path = tmp_path / "fake"

        assert utils.path_exists(file_path) is True
        assert utils.path_exists(fake_path) is False

    def test_rmtree(self, utils: Utils, tmp_path):
        file_path = tmp_path / "root.txt"
        file_path.write_text("foo")
        subdir_path = tmp_path / "dir"
        subdir_path.mkdir()
        subfile_path = subdir_path / "subfile.txt"
        subfile_path.write_text("bar")

        utils.rmtree(tmp_path)
        assert os.path.exists(tmp_path) is False

    @skip_if_windows
    def test_run(self, utils: Utils):
        assert utils.run(["test", "1", "==", "2"]).returncode == 1
        assert utils.run(["test", "1", "==", "1"]).returncode == 0

    @if_windows
    def test_run_windows(self, utils: Utils):
        assert utils.run(["exit", "1"], shell=True).returncode == 1
        assert utils.run(["exit"], shell=True).returncode == 0

    def test_copy_file(self, utils: Utils, tmp_path):
        src_path = tmp_path / "source.txt"
        src_path.write_text("foo")
        dst_path = tmp_path / "destination.txt"

        utils.copy_file(src_path, dst_path)
        with open(src_path) as f:
            src = f.read()
        with open(dst_path) as f:
            dst = f.read()

        assert src == dst

    def test_copy_directory_contents_into(self, utils: Utils, tmp_path):
        src_path = tmp_path / "src"
        src_path.mkdir()
        (src_path / "file").write_text("foo")
        dst_path = tmp_path / "dst"
        dst_path.mkdir()

        utils.copy_directory_contents_into(str(src_path), str(dst_path))

        assert open(dst_path / "file").read() == "foo"

    def test_copy_directory(self, utils: Utils, tmp_path):
        src_path = tmp_path / "src"
        src_path.mkdir()
        (src_path / "file").write_text("foo")
        dst_path = tmp_path / "dst"

        utils.copy_directory(src_path, dst_path)

        assert os.path.exists(dst_path)
        assert open(dst_path / "file").read() == "foo"

    def test_update_metadata(self, utils: Utils, tmp_path):
        data_dir = tmp_path / "awscli" / "data"
        data_dir.mkdir(parents=True)
        metadata_path = data_dir / "metadata.json"
        metadata_path.write_text("{}")

        utils.update_metadata(tmp_path, key="value")

        with open(metadata_path) as f:
            data = json.load(f)
        assert data == {"key": "value"}

    @skip_if_windows
    def test_create_venv(self, utils: Utils, tmp_path):
        utils.create_venv(tmp_path)

        self.assert_dir_has_content(
            tmp_path,
            [
                "bin",
                "include",
                "lib",
            ],
        )
        self.assert_dir_has_content(
            tmp_path / "bin",
            [
                "python",
                "pip",
            ],
        )

    @if_windows
    def test_create_venv_windows(self, utils: Utils, tmp_path):
        utils.create_venv(tmp_path)

        self.assert_dir_has_content(
            tmp_path,
            [
                "Scripts",
                "Include",
                "Lib",
                "pyvenv.cfg",
            ],
        )
        self.assert_dir_has_content(
            tmp_path / "Scripts",
            [
                "python.exe",
                "pip.exe",
            ],
        )

    def assert_dir_has_content(self, path: str, expected_files: List[str]):
        assert set(l.lower() for l in expected_files).issubset(
            set(l.lower() for l in os.listdir(path))
        )


@pytest.fixture
def unmet_error(request):
    error = UnmetDependenciesException(
        [
            ('colorama', '1.0', Requirement('colorama', '>=2.0', '<3.0')),
        ],
        **request.param,
    )
    return str(error)


class TestUnmetDependencies:
    @pytest.mark.parametrize(
        'unmet_error', [{'in_venv': False}], indirect=True
    )
    def test_in_error_message(self, unmet_error):
        assert (
            "colorama (required: ('>=2.0', '<3.0')) (version installed: 1.0)"
        ) in unmet_error
        assert (
            f"{sys.executable} -m pip install --prefer-binary 'colorama>=2.0,<3.0'"
        ) in unmet_error

    @pytest.mark.parametrize(
        'unmet_error', [{'in_venv': False}], indirect=True
    )
    def test_not_in_venv(self, unmet_error):
        assert 'We noticed you are not in a virtualenv.' in unmet_error

    @pytest.mark.parametrize('unmet_error', [{'in_venv': True}], indirect=True)
    def test_in_venv(self, unmet_error):
        assert 'We noticed you are not in a virtualenv.' not in unmet_error

    @pytest.mark.parametrize(
        'unmet_error',
        [{'in_venv': False, 'reason': "custom reason message"}],
        indirect=True,
    )
    def test_custom_reason(self, unmet_error):
        assert 'custom reason message' in unmet_error
